"""
Example demonstrating the attribute inference attacks.

Running
-------

Invoke this code from the root AI-SDC folder with
python -m examples.attribute_inference_example
"""

from __future__ import annotations

import json
import os
import shutil
import sys
import unittest

# ignore unused imports because it depends on whether data file is present
from sklearn.datasets import fetch_openml  # pylint:disable=unused-import
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import (  # pylint:disable=unused-import
    LabelEncoder,
    OneHotEncoder,
)

from aisdc.attacks import attribute_attack  # pylint: disable = import-error
from aisdc.attacks.attribute_attack import (
    _get_bounds_risk,
    _infer_categorical,
    _unique_max,
)
from tests.test_attacks_via_safemodel import get_target

# pylint: disable = duplicate-code


class TestAttributeInferenceAttack(unittest.TestCase):
    """Class which tests the AttributeInferenceAttack module."""

    def _cleanup_file(self, name: str):
        """Removes unwanted files or directory."""
        if os.path.exists(name) and os.path.isfile(name):  # h5
            os.remove(name)
        elif os.path.exists(name) and os.path.isdir(name):  # tf
            shutil.rmtree(name)

    def _common_setup(self):
        """Basic commands to get ready to test some code."""
        model = RandomForestClassifier(bootstrap=False)
        target = get_target(model)
        model.fit(target.x_train, target.y_train)
        attack_obj = attribute_attack.AttributeAttack(n_cpu=7, report_name="aia_report")
        return target, attack_obj

    def test_attack_args(self):
        """Tests methods in the attack_args class."""
        _, attack_obj = self._common_setup()
        attack_obj.__dict__["newkey"] = True
        thedict = attack_obj.__dict__

        self.assertTrue(thedict["newkey"])

    def test_unique_max(self):
        """Tests the _unique_max helper function."""
        has_unique = (0.3, 0.5, 0.2)
        no_unique = (0.5, 0.5)
        self.assertTrue(_unique_max(has_unique, 0.0))
        self.assertFalse(_unique_max(has_unique, 0.6))
        self.assertFalse(_unique_max(no_unique, 0.0))

    def test_categorical_via_modified_attack_brute_force(self):
        """Test lots of functionality for categoricals
        using code from brute_force but without multiprocessing.
        """
        target, _ = self._common_setup()

        threshold = 0
        feature = 0
        # make predictions
        t_low = _infer_categorical(target, feature, threshold)
        t_low_correct = t_low["train"][0]
        t_low_total = t_low["train"][1]
        t_low_train_samples = t_low["train"][4]

        # Check the number of samples in the dataset
        self.assertEqual(len(target.x_train), t_low_train_samples)

        # Check that all samples are correct for this threshold
        self.assertEqual(t_low_correct, t_low_total)

        # or don't because threshold is too high
        threshold = 999
        t_high = _infer_categorical(target, feature, threshold)
        t_high_correct = t_high["train"][0]
        t_high_train_samples = t_high["train"][4]

        self.assertEqual(len(target.x_train), t_high_train_samples)
        self.assertEqual(0, t_high_correct)

    def test_continuous_via_modified_bounds_risk(self):
        """Tests a lot of the code for continuous variables
        via a copy of the _get_bounds_risk()
        modified not to use multiprocessing.
        """
        target, _ = self._common_setup()
        returned = _get_bounds_risk(
            target.model, "dummy", 8, target.x_train, target.x_test
        )

        # Check the number of parameters returned
        self.assertEqual(3, len(returned.keys()))

        # Check the value of the returned parameters
        self.assertEqual(0.0, returned["train"])
        self.assertEqual(0.0, returned["test"])

    # test below covers a lot of the plotting etc.
    def test_AIA_on_nursery(self):
        """Tests running AIA on the nursery data
        with an added continuous feature.
        """
        target, attack_obj = self._common_setup()
        attack_obj.attack(target)

        output = attack_obj.make_report()
        keys = output["attack_experiment_logger"]["attack_instance_logger"][
            "instance_0"
        ].keys()

        self.assertIn("categorical", keys)

    def test_AIA_on_nursery_from_cmd(self):
        """Tests running AIA on the nursery data
        with an added continuous feature.
        """
        target, _ = self._common_setup()
        target.save(path="tests/test_aia_target")

        config = {
            "n_cpu": 7,
            "report_name": "commandline_aia_exampl1_report",
        }
        with open(
            os.path.join("tests", "test_config_aia_cmd.json"), "w", encoding="utf-8"
        ) as f:
            f.write(json.dumps(config))

        cmd_json = os.path.join("tests", "test_config_aia_cmd.json")
        aia_target = os.path.join("tests", "test_aia_target")
        os.system(
            f"{sys.executable} -m aisdc.attacks.attribute_attack run-attack-from-configfile "
            f"--attack-config-json-file-name {cmd_json} "
            f"--attack-target-folder-path {aia_target} "
        )

    def test_cleanup(self):
        """Tidies up any files created."""
        files_made = (
            "delete-me.json",
            "aia_example.json",
            "aia_example.pdf",
            "aia_report_cat_frac.png",
            "aia_report_cat_risk.png",
            "aia_report_quant_risk.png",
            "aia_report.pdf",
            "aia_report.json",
            "aia_attack_from_configfile.json",
            "test_attribute_attack.json",
            os.path.join("tests", "test_config_aia_cmd.json"),
            os.path.join("tests", "test_aia_target/"),
            "output_attribute",
        )
        for fname in files_made:
            self._cleanup_file(fname)



"""
An entry point to run multiple attacks including MIA (worst-case and LIRA)
and attribute inference attack using a single configuration file
with multiple attack configuration.
"""
from __future__ import annotations

import argparse
import json
import logging
import os
import uuid
from typing import Any

from aisdc.attacks.attack import Attack
from aisdc.attacks.attribute_attack import (
    AttributeAttack,  # pylint: disable = import-error
)
from aisdc.attacks.likelihood_attack import LIRAAttack  # pylint: disable = import-error
from aisdc.attacks.target import Target
from aisdc.attacks.worst_case_attack import (
    WorstCaseAttack,  # pylint: disable = import-error
)


class MultipleAttacks(Attack):
    """Class to wrap the MIA and AIA attack codes."""

    def __init__(
        self,
        config_filename: str = None,
    ) -> None:
        super().__init__()
        self.config_filename = config_filename
        """Constructs an object to execute a worst case attack.

        Parameters
        ----------
        config_filename : str
            name of the configuration file which has configurations in a single JSON file
            to support running multiple attacks
        """

    def __str__(self):
        return "Multiple Attacks (MIA and AIA) given configurations"

    def attack(self, target: Target) -> None:
        """
        Runs attacks from a Target object and a target model.

        Parameters
        ----------
        target : attacks.target.Target
            target as an instance of the Target class. Needs to have x_train,
            x_test, y_train and y_test set.
        """
        logger = logging.getLogger("attack-multiple attacks")
        logger.info("Running attacks")
        file_contents = ""
        with open(self.config_filename, "r+", encoding="utf-8") as f:
            file_contents = f.read()

        if file_contents != "":
            config_file_data = json.loads(file_contents)
            for config_obj in config_file_data:
                params = config_file_data[config_obj]
                attack_name = config_obj.split("-")[0]
                attack_obj = None
                if attack_name == "worst_case":
                    attack_obj = WorstCaseAttack(**params)
                elif attack_name == "lira":
                    attack_obj = LIRAAttack(**params)
                elif attack_name == "attribute":
                    attack_obj = AttributeAttack(**params)
                else:
                    attack_names = "'worst_case', 'lira' and 'attribute'"
                    logger.error(
                        """attack name is %s whereas supported attack names are %s: """,
                        attack_name,
                        attack_names,
                    )

                if attack_obj is not None:
                    attack_obj.attack(target)

                if attack_obj is not None:
                    _ = attack_obj.make_report()
        logger.info("Finished running attacks")


class ConfigFile:  # pylint: disable = too-few-public-methods
    """Module that creates a single JSON configuration file."""

    def __init__(
        self,
        filename: str = None,
    ) -> None:
        self.filename = filename

        dirname = os.path.normpath(os.path.dirname(self.filename))
        os.makedirs(dirname, exist_ok=True)
        # if file doesn't exist, create it
        with open(self.filename, "w", encoding="utf-8") as f:
            f.write("")

    def add_config(self, config_obj: Any, config_attack_type: str) -> None:
        """Add a section of JSON to the file which is already open."""

        # Read the contents of the file and then clear the file
        config_file_data = self.read_config_file()

        # Add the new JSON to the JSON that was in the file, and re-write
        with open(self.filename, "w", encoding="utf-8") as f:
            class_name = config_attack_type + "-" + str(uuid.uuid4())

            if isinstance(config_obj, dict):
                config_file_data[class_name] = config_obj
            elif isinstance(config_obj, str):
                with open(str(config_obj), encoding="utf-8") as fr:
                    config_file_data[class_name] = json.loads(fr.read())

            f.write(json.dumps(config_file_data))

    def read_config_file(self) -> dict:
        """Reads a JSON configuration file and returns dictionary
        with a number of configuration objects.
        """
        with open(self.filename, encoding="utf-8") as f:
            file_contents = f.read()
            if file_contents != "":
                config_file_data = json.loads(file_contents)
            else:
                config_file_data = {}
        return config_file_data


def _run_attack_from_configfile(args):
    """Run a command line attack based on saved files described in .json file."""
    attack_obj = MultipleAttacks(
        config_filename=str(args.config_filename),
    )
    target = Target()
    target.load(args.target_path)
    attack_obj.attack(target)


def main():
    """Main method to parse args and invoke relevant code."""
    parser = argparse.ArgumentParser(add_help=False)

    subparsers = parser.add_subparsers()
    attack_parser_config = subparsers.add_parser("run-attack-from-configfile")
    attack_parser_config.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="config_filename",
        type=str,
        default="singleconfig.json",
        help=(
            """Name of the .json file containing details for running
            multiple attacks run. Default = %(default)s"""
        ),
    )

    attack_parser_config.add_argument(
        "-t",
        "--attack-target-folder-path",
        action="store",
        required=True,
        dest="target_path",
        type=str,
        default="target",
        help=(
            """Name of the target directory to load the trained target model and the target data.
        Default = %(default)s"""
        ),
    )

    attack_parser_config.set_defaults(func=_run_attack_from_configfile)
    args = parser.parse_args()
    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        print(e)
        print("Invalid command. Try --help to get more details")


if __name__ == "__main__":  # pragma:no cover
    main()



"""
Worst_case_attack.py.

Runs a worst case attack based upon predictive probabilities stored in two .csv files
"""  # pylint: disable = too-many-lines

from __future__ import annotations

import argparse
import logging
import os
import uuid
from datetime import datetime
from typing import Any, Iterable

import numpy as np
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split

from aisdc import metrics
from aisdc.attacks import report
from aisdc.attacks.attack import Attack
from aisdc.attacks.attack_report_formatter import GenerateJSONModule
from aisdc.attacks.failfast import FailFast
from aisdc.attacks.target import Target

logging.basicConfig(level=logging.INFO)

P_THRESH = 0.05


class WorstCaseAttack(Attack):
    """Class to wrap the worst case attack code."""

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable = too-many-arguments, too-many-locals, too-many-statements
        self,
        n_reps: int = 10,
        reproduce_split: int | Iterable[int] | None = 5,
        p_thresh: float = 0.05,
        n_dummy_reps: int = 1,
        train_beta: int = 1,
        test_beta: int = 1,
        test_prop: float = 0.2,
        n_rows_in: int = 1000,
        n_rows_out: int = 1000,
        training_preds_filename: str = None,
        test_preds_filename: str = None,
        output_dir: str = "output_worstcase",
        report_name: str = "report_worstcase",
        include_model_correct_feature: bool = False,
        sort_probs: bool = True,
        mia_attack_model: Any = RandomForestClassifier,
        mia_attack_model_hyp: dict = None,
        attack_metric_success_name: str = "P_HIGHER_AUC",
        attack_metric_success_thresh: float = 0.05,
        attack_metric_success_comp_type: str = "lte",
        attack_metric_success_count_thresh: int = 5,
        attack_fail_fast: bool = False,
        attack_config_json_file_name: str = None,
        target_path: str = None,
    ) -> None:
        """Constructs an object to execute a worst case attack.

        Parameters
        ----------
        n_reps : int
            number of attacks to run -- in each iteration an attack model
            is trained on a different subset of the data
        reproduce_split : int
            variable that controls the reproducibility of the data split.
            It can be an integer or a list of integers of length n_reps. Default : 5.
        p_thresh : float
            threshold to determine significance of things. For instance auc_p_value and pdif_vals
        n_dummy_reps : int
            number of baseline (dummy) experiments to do
        train_beta : int
            value of b for beta distribution used to sample the in-sample (training) probabilities
        test_beta : int
            value of b for beta distribution used to sample the out-of-sample (test) probabilities
        test_prop : float
            proportion of data to use as a test set for the attack model
        n_rows_in : int
            number of rows for in-sample (training data)
        n_rows_out : int
            number of rows for out-of-sample (test data)
        training_preds_filename : str
            name of the file to keep predictions of the training data (in-sample)
        test_preds_filename : str
            name of the file to keep predictions of the test data (out-of-sample)
        output_dir : str
            name of the directory where outputs are stored
        report_name : str
            name of the pdf and json output reports
        include_model_correct_feature : bool
            inclusion of additional feature to hold whether or not the target model
            made a correct prediction for each example
        sort_probs : bool
            true in case require to sort combine preds (from training and test)
            to have highest probabilities in the first column
        mia_attack_model : Any
            name of the attack model such as RandomForestClassifier
        mia_attack_model_hyp : dict
            dictionary of hyper parameters for the mia_attack_model
            such as min_sample_split, min_samples_leaf etc
        attack_metric_success_name : str
            name of metric to compute for the attack being successful
        attack_metric_success_thresh : float
            threshold for a given metric to measure attack being successful or not
        attack_metric_success_comp_type : str
            threshold comparison operator (i.e., gte: greater than or equal to, gt:
            greater than, lte: less than or equal to, lt: less than,
            eq: equal to and not_eq: not equal to)
        attack_metric_success_count_thresh : int
            a counter to record how many times an attack was successful
            given that the threshold has fulfilled criteria for a given comparison type
        attack_fail_fast : bool
            If true it stops repetitions earlier based on the given attack metric
            (i.e., attack_metric_success_name) considering the comparison type
            (attack_metric_success_comp_type) satisfying a threshold
            (i.e., attack_metric_success_thresh) for n
            (attack_metric_success_count_thresh) number of times
        attack_config_json_file_name : str
            name of the configuration file to load parameters
        target_path : str
            path to the saved trained target model and target data
        """

        super().__init__()
        self.n_reps = n_reps
        self.reproduce_split = reproduce_split
        if isinstance(reproduce_split, int):
            reproduce_split = [reproduce_split] + [
                x**2 for x in range(reproduce_split, reproduce_split + n_reps - 1)
            ]  # pylint: disable = line-too-long
        else:
            reproduce_split = list(
                dict.fromkeys(reproduce_split)
            )  # remove potential duplicates
            if len(reproduce_split) == n_reps:
                pass
            elif len(reproduce_split) > n_reps:
                print("split", reproduce_split, "nreps", n_reps)
                reproduce_split = list(reproduce_split)[0:n_reps]
                print(
                    "WARNING: the length of the parameter 'reproduce_split'\
                     is longer than n_reps. Values have been removed."
                )
            else:
                # assign values to match length of n_reps
                reproduce_split += [
                    reproduce_split[-1] * x
                    for x in range(2, (n_reps - len(reproduce_split) + 2))
                ]
                print(
                    "WARNING: the length of the parameter 'reproduce_split'\
                     is shorter than n_reps. Vales have been added."
                )
            print("reproduce split now", reproduce_split)
        self.reproduce_split = reproduce_split
        self.p_thresh = p_thresh
        self.n_dummy_reps = n_dummy_reps
        self.train_beta = train_beta
        self.test_beta = test_beta
        self.test_prop = test_prop
        self.n_rows_in = n_rows_in
        self.n_rows_out = n_rows_out
        self.training_preds_filename = training_preds_filename
        self.test_preds_filename = test_preds_filename
        self.output_dir = output_dir
        self.report_name = report_name
        self.include_model_correct_feature = include_model_correct_feature
        self.sort_probs = sort_probs
        self.mia_attack_model = mia_attack_model
        if mia_attack_model_hyp is None:
            self.mia_attack_model_hyp = {
                "min_samples_split": 20,
                "min_samples_leaf": 10,
                "max_depth": 5,
            }
        else:
            self.mia_attack_model_hyp = mia_attack_model_hyp
        self.attack_metric_success_name = attack_metric_success_name
        self.attack_metric_success_thresh = attack_metric_success_thresh
        self.attack_metric_success_comp_type = attack_metric_success_comp_type
        self.attack_metric_success_count_thresh = attack_metric_success_count_thresh
        self.attack_fail_fast = attack_fail_fast
        self.attack_config_json_file_name = attack_config_json_file_name
        self.target_path = target_path
        # Updating parameters from a configuration json file
        if self.attack_config_json_file_name is not None:
            self._update_params_from_config_file()
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        self.attack_metrics = None
        self.attack_metric_failfast_summary = None
        self.dummy_attack_metrics = None
        self.dummy_attack_metric_failfast_summary = None
        self.metadata = None

    def __str__(self):
        return "WorstCase attack"

    def attack(self, target: Target) -> None:
        """Programmatic attack entry point.

        To be used when code has access to Target class and trained target model

        Parameters
        ----------
        target : attacks.target.Target
            target as a Target class object
        """
        train_preds = target.model.predict_proba(target.x_train)
        test_preds = target.model.predict_proba(target.x_test)
        train_correct = None
        test_correct = None
        if self.include_model_correct_feature:
            train_correct = 1 * (target.y_train == target.model.predict(target.x_train))
            test_correct = 1 * (target.y_test == target.model.predict(target.x_test))

        self.attack_from_preds(
            train_preds,
            test_preds,
            train_correct=train_correct,
            test_correct=test_correct,
        )

    def attack_from_prediction_files(self):
        """Start an attack from saved prediction files.

        To be used when only saved predictions are available.

        Filenames for the saved prediction files to be specified in the arguments provided
        in the constructor
        """
        train_preds = np.loadtxt(self.training_preds_filename, delimiter=",")
        test_preds = np.loadtxt(self.test_preds_filename, delimiter=",")
        self.attack_from_preds(train_preds, test_preds)

    def attack_from_preds(  # pylint: disable=too-many-locals
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> None:
        """
        Runs the attack based upon the predictions in train_preds and test_preds, and the params
        stored in self.args.

        Parameters
        ----------
        train_preds : np.ndarray
            Array of train predictions. One row per example, one column per class (i.e. 2)
        test_preds : np.ndarray
            Array of test predictions. One row per example, one column per class (i.e. 2)
        """
        logger = logging.getLogger("attack-from-preds")
        logger.info("Running main attack repetitions")
        attack_metric_dict = self.run_attack_reps(
            train_preds,
            test_preds,
            train_correct=train_correct,
            test_correct=test_correct,
        )
        self.attack_metrics = attack_metric_dict["mia_metrics"]
        self.attack_metric_failfast_summary = attack_metric_dict[
            "failfast_metric_summary"
        ]

        self.dummy_attack_metrics = []
        self.dummy_attack_metric_failfast_summary = []
        if self.n_dummy_reps > 0:
            logger.info("Running dummy attack reps")
            n_train_rows = len(train_preds)
            n_test_rows = len(test_preds)
            for _ in range(self.n_dummy_reps):
                d_train_preds, d_test_preds = self.generate_arrays(
                    n_train_rows,
                    n_test_rows,
                    self.train_beta,
                    self.test_beta,
                )
                temp_attack_metric_dict = self.run_attack_reps(
                    d_train_preds, d_test_preds
                )
                temp_metrics = temp_attack_metric_dict["mia_metrics"]
                temp_metric_failfast_summary = temp_attack_metric_dict[
                    "failfast_metric_summary"
                ]

                self.dummy_attack_metrics.append(temp_metrics)
                self.dummy_attack_metric_failfast_summary.append(
                    temp_metric_failfast_summary
                )

        logger.info("Finished running attacks")

    def _prepare_attack_data(
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Prepare training data and labels for attack model
        Combines the train and test preds into a single numpy array (optionally) sorting each
        row to have the highest probabilities in the first column. Constructs a label array that
        has ones corresponding to training rows and zeros to testing rows.
        """
        logger = logging.getLogger("prep-attack-data")
        if self.sort_probs:
            logger.info("Sorting probabilities to leave highest value in first column")
            train_preds = -np.sort(-train_preds, axis=1)
            test_preds = -np.sort(-test_preds, axis=1)

        logger.info("Creating MIA data")

        if self.include_model_correct_feature and train_correct is not None:
            train_preds = np.hstack((train_preds, train_correct[:, None]))
            test_preds = np.hstack((test_preds, test_correct[:, None]))

        mi_x = np.vstack((train_preds, test_preds))
        mi_y = np.hstack((np.ones(len(train_preds)), np.zeros(len(test_preds))))

        return (mi_x, mi_y)

    def run_attack_reps(  # pylint: disable = too-many-locals
        self,
        train_preds: np.ndarray,
        test_preds: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> dict:
        """
        Run actual attack reps from train and test predictions.

        Parameters
        ----------
        train_preds : np.ndarray
            predictions from the model on training (in-sample) data
        test_preds : np.ndarray
            predictions from the model on testing (out-of-sample) data

        Returns
        -------
        mia_metrics_dict : dict
            a dictionary with two items including mia_metrics
            (a list of metric across repetitions) and failfast_metric_summary object
            (an object of FailFast class) to maintain summary of
            fail/success of attacks for a given metric of failfast option
        """
        self.n_rows_in = len(train_preds)
        self.n_rows_out = len(test_preds)
        logger = logging.getLogger("attack-reps")
        mi_x, mi_y = self._prepare_attack_data(
            train_preds, test_preds, train_correct, test_correct
        )

        mia_metrics = []

        failfast_metric_summary = FailFast(self)

        for rep in range(self.n_reps):
            logger.info(
                "Rep %d of %d split %d", rep + 1, self.n_reps, self.reproduce_split[rep]
            )
            mi_train_x, mi_test_x, mi_train_y, mi_test_y = train_test_split(
                mi_x,
                mi_y,
                test_size=self.test_prop,
                stratify=mi_y,
                random_state=self.reproduce_split[rep],
                shuffle=True,
            )
            attack_classifier = self.mia_attack_model(**self.mia_attack_model_hyp)
            attack_classifier.fit(mi_train_x, mi_train_y)
            y_pred_proba, y_test = metrics.get_probabilities(
                attack_classifier, mi_test_x, mi_test_y, permute_rows=True
            )

            mia_metrics.append(metrics.get_metrics(y_pred_proba, y_test))

            if self.include_model_correct_feature and train_correct is not None:
                # Compute the Yeom TPR and FPR
                yeom_preds = mi_test_x[:, -1]
                tn, fp, fn, tp = confusion_matrix(mi_test_y, yeom_preds).ravel()
                mia_metrics[-1]["yeom_tpr"] = tp / (tp + fn)
                mia_metrics[-1]["yeom_fpr"] = fp / (fp + tn)
                mia_metrics[-1]["yeom_advantage"] = (
                    mia_metrics[-1]["yeom_tpr"] - mia_metrics[-1]["yeom_fpr"]
                )

            failfast_metric_summary.check_attack_success(mia_metrics[rep])

            if (
                failfast_metric_summary.check_overall_attack_success(self)
                and self.attack_fail_fast
            ):
                break

        logger.info("Finished simulating attacks")

        mia_metrics_dict = {}
        mia_metrics_dict["mia_metrics"] = mia_metrics
        mia_metrics_dict["failfast_metric_summary"] = failfast_metric_summary

        return mia_metrics_dict

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Summarise metrics from a metric list.

        Returns
        -------
        global_metrics : Dict
            Dictionary of summary metrics

        Arguments
        ---------
        attack_metrics: List
            list of attack metrics dictionaries
        """
        global_metrics = {}
        if attack_metrics is not None and len(attack_metrics) != 0:
            auc_p_vals = [
                metrics.auc_p_val(
                    m["AUC"], m["n_pos_test_examples"], m["n_neg_test_examples"]
                )[0]
                for m in attack_metrics
            ]

            m = attack_metrics[0]
            _, auc_std = metrics.auc_p_val(
                0.5, m["n_pos_test_examples"], m["n_neg_test_examples"]
            )

            global_metrics[
                "null_auc_3sd_range"
            ] = f"{0.5 - 3*auc_std:.4f} -> {0.5 + 3*auc_std:.4f}"
            global_metrics["n_sig_auc_p_vals"] = self._get_n_significant(
                auc_p_vals, self.p_thresh
            )
            global_metrics["n_sig_auc_p_vals_corrected"] = self._get_n_significant(
                auc_p_vals, self.p_thresh, bh_fdr_correction=True
            )

            pdif_vals = [np.exp(-m["PDIF01"]) for m in attack_metrics]
            global_metrics["n_sig_pdif_vals"] = self._get_n_significant(
                pdif_vals, self.p_thresh
            )
            global_metrics["n_sig_pdif_vals_corrected"] = self._get_n_significant(
                pdif_vals, self.p_thresh, bh_fdr_correction=True
            )

        return global_metrics

    def _get_n_significant(self, p_val_list, p_thresh, bh_fdr_correction=False):
        """
        Helper method to determine if values within a list of p-values are significant at
        p_thresh. Can perform multiple testing correction.
        """
        if not bh_fdr_correction:
            return sum(
                1 for p in p_val_list if p <= p_thresh
            )  # pylint: disable = consider-using-generator
        p_val_list = np.asarray(sorted(p_val_list))
        n_vals = len(p_val_list)
        hoch_vals = np.array([(k / n_vals) * P_THRESH for k in range(1, n_vals + 1)])
        bh_sig_list = p_val_list <= hoch_vals
        if any(bh_sig_list):
            n_sig_bh = (np.where(bh_sig_list)[0]).max() + 1
        else:
            n_sig_bh = 0
        return n_sig_bh

    def _generate_array(self, n_rows: int, beta: float) -> np.ndarray:
        """Generate a single array of predictions, used when doing baseline experiments.

        Parameters
        ----------
        n_rows : int
            the number of rows worth of data to generate
        beta : float
            the beta parameter for sampling probabilities

        Returns
        -------
        preds : np.ndarray
            Array of predictions. Two columns, n_rows rows

        Notes
        -----

        Examples
        --------
        """

        preds = np.zeros((n_rows, 2), float)
        for row_idx in range(n_rows):
            train_class = np.random.choice(2)
            train_prob = np.random.beta(1, beta)
            preds[row_idx, train_class] = train_prob
            preds[row_idx, 1 - train_class] = 1 - train_prob
        return preds

    def generate_arrays(
        self,
        n_rows_in: int,
        n_rows_out: int,
        train_beta: float = 2,
        test_beta: float = 2,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Generate train and test prediction arrays, used when computing baseline.

        Parameters
        ----------
        n_rows_in : int
            number of rows of in-sample (training) probabilities
        n_rows_out : int
            number of rows of out-of-sample (testing) probabilities
        train_beta : float
            beta value for generating train probabilities
        test_beta : float:
            beta_value for generating test probabilities

        Returns
        -------
        train_preds : np.ndarray
            Array of train predictions (n_rows x 2 columns)
        test_preds : np.ndarray
            Array of test predictions (n_rows x 2 columns)
        """
        train_preds = self._generate_array(n_rows_in, train_beta)
        test_preds = self._generate_array(n_rows_out, test_beta)
        return train_preds, test_preds

    def make_dummy_data(self) -> None:
        """Makes dummy data for testing functionality.

        Parameters
        ----------
        args : dict
            Command line arguments

        Returns
        -------

        Notes
        -----
        Returns nothing but saves two .csv files
        """
        logger = logging.getLogger("dummy-data")
        logger.info(
            "Making dummy data with %d rows in and %d out",
            self.n_rows_in,
            self.n_rows_out,
        )
        logger.info("Generating rows")
        train_preds, test_preds = self.generate_arrays(
            self.n_rows_in,
            self.n_rows_out,
            train_beta=self.train_beta,
            test_beta=self.test_beta,
        )
        logger.info("Saving files")
        np.savetxt(self.training_preds_filename, train_preds, delimiter=",")
        np.savetxt(self.test_preds_filename, test_preds, delimiter=",")

    def _construct_metadata(self):
        """Constructs the metadata object, after attacks."""
        self.metadata = {}
        # Store all args
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"] = self.get_params()

        self.metadata["attack"] = str(self)

        # Global metrics
        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)
        self.metadata["baseline_global_metrics"] = self._get_global_metrics(
            self._unpack_dummy_attack_metrics_experiments_instances()
        )

    def _unpack_dummy_attack_metrics_experiments_instances(self) -> list:
        """Constructs the metadata object, after attacks."""
        dummy_attack_metrics_instances = []

        for exp_rep, _ in enumerate(self.dummy_attack_metrics):
            temp_dummy_attack_metrics = self.dummy_attack_metrics[exp_rep]
            dummy_attack_metrics_instances += temp_dummy_attack_metrics

        return dummy_attack_metrics_instances

    def _get_attack_metrics_instances(self) -> dict:
        """Constructs the metadata object, after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}

        for rep, _ in enumerate(self.attack_metrics):
            attack_metrics_instances["instance_" + str(rep)] = self.attack_metrics[rep]

        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        attack_metrics_experiment[
            "attack_metric_failfast_summary"
        ] = self.attack_metric_failfast_summary.get_attack_summary()

        return attack_metrics_experiment

    def _get_dummy_attack_metrics_experiments_instances(self) -> dict:
        """Constructs the metadata object, after attacks."""
        dummy_attack_metrics_experiments = {}

        for exp_rep, _ in enumerate(self.dummy_attack_metrics):
            temp_dummy_attack_metrics = self.dummy_attack_metrics[exp_rep]
            dummy_attack_metric_instances = {}
            for rep, _ in enumerate(temp_dummy_attack_metrics):
                dummy_attack_metric_instances[
                    "instance_" + str(rep)
                ] = temp_dummy_attack_metrics[rep]
            temp = {}
            temp["attack_instance_logger"] = dummy_attack_metric_instances
            temp[
                "attack_metric_failfast_summary"
            ] = self.dummy_attack_metric_failfast_summary[exp_rep].get_attack_summary()
            dummy_attack_metrics_experiments[
                "dummy_attack_metrics_experiment_" + str(exp_rep)
            ] = temp

        return dummy_attack_metrics_experiments

    def make_report(self) -> dict:
        """Creates output dictionary structure and generates
        pdf and json outputs if filenames are given.
        """
        output = {}
        output["log_id"] = str(uuid.uuid4())
        output["log_time"] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

        self._construct_metadata()
        output["metadata"] = self.metadata

        output["attack_experiment_logger"] = self._get_attack_metrics_instances()
        output[
            "dummy_attack_experiments_logger"
        ] = self._get_dummy_attack_metrics_experiments_instances()

        report_dest = os.path.join(self.output_dir, self.report_name)
        json_attack_formatter = GenerateJSONModule(report_dest + ".json")
        json_report = report.create_json_report(output)
        json_attack_formatter.add_attack_output(json_report, "WorstCaseAttack")

        pdf_report = report.create_mia_report(output)
        report.add_output_to_pdf(report_dest, pdf_report, "WorstCaseAttack")
        return output


def _make_dummy_data(args):
    """Initialise class and run dummy data creation."""
    args.__dict__["training_preds_filename"] = "train_preds.csv"
    args.__dict__["test_preds_filename"] = "test_preds.csv"
    attack_obj = WorstCaseAttack(
        train_beta=args.train_beta,
        test_beta=args.test_beta,
        n_rows_in=args.n_rows_in,
        n_rows_out=args.n_rows_out,
        training_preds_filename=args.training_preds_filename,
        test_preds_filename=args.test_preds_filename,
    )
    attack_obj.make_dummy_data()


def _run_attack(args):
    """Initialise class and run attack from prediction files."""
    attack_obj = WorstCaseAttack(
        n_reps=args.n_reps,
        p_thresh=args.p_thresh,
        n_dummy_reps=args.n_dummy_reps,
        train_beta=args.train_beta,
        test_beta=args.test_beta,
        test_prop=args.test_prop,
        training_preds_filename=args.training_preds_filename,
        test_preds_filename=args.test_preds_filename,
        output_dir=args.output_dir,
        report_name=args.report_name,
        sort_probs=args.sort_probs,
        attack_metric_success_name=args.attack_metric_success_name,
        attack_metric_success_thresh=args.attack_metric_success_thresh,
        attack_metric_success_comp_type=args.attack_metric_success_comp_type,
        attack_metric_success_count_thresh=args.attack_metric_success_count_thresh,
        attack_fail_fast=args.attack_fail_fast,
    )
    print(attack_obj.training_preds_filename)
    attack_obj.attack_from_prediction_files()
    _ = attack_obj.make_report()


def _run_attack_from_configfile(args):
    """Initialise class and run attack from prediction files using config file."""
    attack_obj = WorstCaseAttack(
        attack_config_json_file_name=str(args.attack_config_json_file_name),
        target_path=str(args.target_path),
    )
    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    _ = attack_obj.make_report()


def main():
    """Main method to parse arguments and invoke relevant method."""
    logger = logging.getLogger("main")
    parser = argparse.ArgumentParser(
        description=("Perform a worst case attack from saved model predictions")
    )

    subparsers = parser.add_subparsers()
    dummy_parser = subparsers.add_parser("make-dummy-data")
    dummy_parser.add_argument(
        "--num-rows-in",
        action="store",
        dest="n_rows_in",
        type=int,
        required=False,
        default=1000,
        help=("How many rows to generate in the in-sample file. Default = %(default)d"),
    )

    dummy_parser.add_argument(
        "--num-rows-out",
        action="store",
        dest="n_rows_out",
        type=int,
        required=False,
        default=1000,
        help=(
            "How many rows to generate in the out-of-sample file. Default = %(default)d"
        ),
    )

    dummy_parser.add_argument(
        "--train-beta",
        action="store",
        type=float,
        required=False,
        default=5,
        dest="train_beta",
        help=(
            """Value of b parameter for beta distribution used to sample the in-sample
            probabilities. High values will give more extreme probabilities. Set this
            value higher than --test-beta to see successful attacks. Default = %(default)f"""
        ),
    )

    dummy_parser.add_argument(
        "--test-beta",
        action="store",
        type=float,
        required=False,
        default=2,
        dest="test_beta",
        help=(
            "Value of b parameter for beta distribution used to sample the out-of-sample "
            "probabilities. High values will give more extreme probabilities. Set this value "
            "lower than --train-beta to see successful attacks. Default = %(default)f"
        ),
    )

    dummy_parser.set_defaults(func=_make_dummy_data)

    attack_parser = subparsers.add_parser("run-attack")
    attack_parser.add_argument(
        "-i",
        "--training-preds-filename",
        action="store",
        dest="training_preds_filename",
        required=False,
        type=str,
        default="train_preds.csv",
        help=(
            "csv file containing the predictive probabilities (one column per class) for the "
            "training data (one row per training example). Default = %(default)s"
        ),
    )

    attack_parser.add_argument(
        "-o",
        "--test-preds-filename",
        action="store",
        dest="test_preds_filename",
        required=False,
        type=str,
        default="test_preds.csv",
        help=(
            "csv file containing the predictive probabilities (one column per class) for the "
            "non-training data (one row per training example). Default = %(default)s"
        ),
    )

    attack_parser.add_argument(
        "-r",
        "--n-reps",
        type=int,
        required=False,
        default=5,
        action="store",
        dest="n_reps",
        help=(
            "Number of repetitions (splitting data into attack model training and testing "
            "partitions to perform. Default = %(default)d"
        ),
    )

    attack_parser.add_argument(
        "-t",
        "--test-prop",
        type=float,
        required=False,
        default=0.3,
        action="store",
        dest="test_prop",
        help=(
            "Proportion of examples to be used for testing when fiting the attack model. "
            "Default = %(default)f"
        ),
    )

    attack_parser.add_argument(
        "--output-dir",
        type=str,
        action="store",
        dest="output_dir",
        default="output_worstcase",
        required=False,
        help=("Directory name where output files are stored. Default = %(default)s."),
    )

    attack_parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        default="report_worstcase",
        required=False,
        help=(
            """Filename for the pdf and json report outputs. Default = %(default)s.
            Code will append .pdf and .json"""
        ),
    )

    attack_parser.add_argument(
        "--n-dummy-reps",
        type=int,
        action="store",
        dest="n_dummy_reps",
        default=1,
        required=False,
        help=(
            "Number of dummy datasets to sample. Each will be assessed with --n-reps train and "
            "test splits. Set to 0 to do no baseline calculations. Default = %(default)d"
        ),
    )

    attack_parser.add_argument(
        "--p-thresh",
        action="store",
        type=float,
        default=P_THRESH,
        required=False,
        dest="p_thresh",
        help=("P-value threshold for significance testing. Default = %(default)f"),
    )

    attack_parser.add_argument(
        "--train-beta",
        action="store",
        type=float,
        required=False,
        default=5,
        dest="train_beta",
        help=(
            "Value of b parameter for beta distribution used to sample the in-sample probabilities."
            "High values will give more extreme probabilities. Set this value higher than "
            "--test-beta to see successful attacks. Default = %(default)f"
        ),
    )

    attack_parser.add_argument(
        "--test-beta",
        action="store",
        type=float,
        required=False,
        default=2,
        dest="test_beta",
        help=(
            "Value of b parameter for beta distribution used to sample the out-of-sample "
            "probabilities. High values will give more extreme probabilities. Set this value "
            "lower than --train-beta to see successful attacks. Default = %(default)f"
        ),
    )

    # --include-correct feature not supported as not currently possible from the command line
    # as we cannot compute the correctness of predictions.

    attack_parser.add_argument(
        "--sort-probs",
        action="store",
        type=bool,
        default=True,
        required=False,
        dest="sort_probs",
        help=(
            "Whether or not to sort the output probabilities (per row) before "
            "using them to train the attack model. Default = %(default)f"
        ),
    )

    attack_parser.add_argument(
        "--attack-metric-success-name",
        action="store",
        type=str,
        default="P_HIGHER_AUC",
        required=False,
        dest="attack_metric_success_name",
        help=(
            """for computing attack success/failure based on
            --attack-metric-success-thresh option. Default = %(default)s"""
        ),
    )

    attack_parser.add_argument(
        "--attack-metric-success-thresh",
        action="store",
        type=float,
        default=0.05,
        required=False,
        dest="attack_metric_success_thresh",
        help=(
            """for defining threshold value to measure attack success
            for the metric defined by argument --fail-metric-name option. Default = %(default)f"""
        ),
    )

    attack_parser.add_argument(
        "--attack-metric-success-comp-type",
        action="store",
        type=str,
        default="lte",
        required=False,
        dest="attack_metric_success_comp_type",
        help=(
            """for computing attack success/failure based on
            --attack-metric-success-thresh option. Default = %(default)s"""
        ),
    )

    attack_parser.add_argument(
        "--attack-metric-success-count-thresh",
        action="store",
        type=int,
        default=2,
        required=False,
        dest="attack_metric_success_count_thresh",
        help=(
            """for setting counter limit to stop further repetitions given the attack is
             successful and the --attack-fail-fast is true. Default = %(default)d"""
        ),
    )

    attack_parser.add_argument(
        "--attack-fail-fast",
        action="store_true",
        required=False,
        dest="attack_fail_fast",
        help=(
            """to stop further repetitions when the given metric has fulfilled
            a criteria for a specified number of times (--attack-metric-success-count-thresh)
            and this has a true status. Default = %(default)s"""
        ),
    )

    attack_parser.set_defaults(func=_run_attack)

    attack_parser_config = subparsers.add_parser("run-attack-from-configfile")
    attack_parser_config.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="attack_config_json_file_name",
        type=str,
        default="config_worstcase_cmd.json",
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )

    attack_parser_config.add_argument(
        "-t",
        "--attack-target-folder-path",
        action="store",
        required=True,
        dest="target_path",
        type=str,
        default="worstcase_target",
        help=(
            """Name of the target directory to load the trained target model and the target data.
            Default = %(default)s"""
        ),
    )

    attack_parser_config.set_defaults(func=_run_attack_from_configfile)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        logger.error("Invalid command. Try --help to get more details")
        logger.error(e)


if __name__ == "__main__":  # pragma:no cover
    main()



"""Test_generate_report.py
Copyright (C) Jim Smith 2022 <james.smith@uwe.ac.uk>.
"""

from __future__ import annotations

import json
import os
import shutil
import unittest

import pytest

from aisdc.attacks.attack_report_formatter import (
    AnalysisModule,
    FinalRecommendationModule,
    GenerateJSONModule,
    GenerateTextReport,
    LogLogROCModule,
    SummariseAUCPvalsModule,
    SummariseFDIFPvalsModule,
    SummariseUnivariateMetricsModule,
)


def get_test_report():
    """Create a mock attack result dictionary for use with tests."""
    json_formatted = {
        "log_id": 1024,
        "metadata": {"attack": "WorstCase"},
        "model_params": {"min_samples_leaf": 10},
        "model_name": "RandomForestClassifier",
        "WorstCaseAttack": {"attack_experiment_logger": {"attack_instance_logger": {}}},
    }

    metrics_dict = {
        "P_HIGHER_AUC": 0.5,
        "AUC": 0.6,
        "ACC": 0.6,
        "FDIF01": 0.2,
        "PDIF01": 1.0,
        "FPR": 1.0,
        "TPR": 0.1,
        "fpr": [0.0, 0.0, 1.0],
        "tpr": [0.0, 1.0, 1.0],
    }

    for i in range(10):
        json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
            "attack_instance_logger"
        ]["instance_" + str(i)] = metrics_dict

    return json_formatted


def get_target_report():
    """Create a mock target model dictionary for use with tests."""
    target_formatted = {
        "data_name": "",
        "n_samples": 12960,
        "features": {},
        "n_features": 0,
        "n_samples_orig": 0,
        "model_path": "model.pkl",
        "model_name": "SVC",
        "model_params": {"C": 1.0},
    }

    return target_formatted


def clean_up(name):
    """Removes unwanted files or directory."""
    if os.path.exists(name) and os.path.isfile(name):  # h5
        os.remove(name)
    elif os.path.exists(name) and os.path.isdir(name):  # tf
        shutil.rmtree(name)


class TestGenerateReport(unittest.TestCase):
    """Class which tests the attack_report_formatter.py file."""

    def process_json_from_file(self, json_formatted):
        """Function which handles file input/output from the process_json function."""
        filename = "test.json"
        output_filename = "results.txt"

        with open(filename, "w", encoding="utf-8") as f:
            json.dump(json_formatted, f)

        g = GenerateTextReport()
        g.process_attack_target_json(filename)
        g.export_to_file(output_filename)

        with open(output_filename, encoding="utf-8") as file:
            data = file.read()

        clean_up(filename)
        clean_up(output_filename)

        return data

    def test_not_implemented(self):
        """Test to make sure analysis module fails expectedly when functions are called directly."""
        a = AnalysisModule()
        with pytest.raises(NotImplementedError):
            a.process_dict()

        with pytest.raises(NotImplementedError):
            str(a)

    def test_json_formatter(self):
        """Test which tests the GenerateJSONModule."""
        g = GenerateJSONModule()
        filename = g.get_output_filename()
        self.assertIsNotNone(filename)
        clean_up(filename)

        test_filename = "example_filename.json"
        g = GenerateJSONModule(test_filename)
        self.assertEqual(test_filename, g.get_output_filename())
        clean_up(test_filename)

        # check file is overwritten when the same file is passed
        test_filename = "filename_to_rewrite.json"
        g = GenerateJSONModule(test_filename)
        g.clean_file()

        msg_1 = "this should be included in the file"
        msg_2 = "this should also be included in the file"

        g.add_attack_output('{"test_output":"' + msg_1 + '"}', "FirstTestAttack")
        g.add_attack_output('{"test_output":"' + msg_2 + '"}', "SecondTestAttack")

        with open(test_filename, encoding="utf-8") as f:
            file_contents = json.loads(f.read())

        self.assertIn(msg_1, file_contents["FirstTestAttack"]["test_output"])
        self.assertIn(msg_2, file_contents["SecondTestAttack"]["test_output"])

        clean_up(test_filename)

    def test_process_attack_target_json(self):
        """Test which tests the process_attack_target_json function."""
        target_report = get_target_report()
        target_json = "target.json"

        with open(target_json, "w", encoding="utf-8") as f:
            json.dump(target_report, f)

        json_formatted = get_test_report()

        attack_json = "test.json"
        output_filename = "attack.txt"

        with open(attack_json, "w", encoding="utf-8") as f:
            json.dump(json_formatted, f)

        g = GenerateTextReport()
        g.process_attack_target_json(attack_json, target_json)
        g.export_to_file(output_filename)

        metrics_dict = {
            "P_HIGHER_AUC": 0.001,
            "AUC": 0.8,
            "ACC": 0.7,
            "FDIF01": 0.2,
            "PDIF01": 1.0,
            "FPR": 1.0,
            "TPR": 0.1,
        }

        for i in range(10):
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
                "attack_instance_logger"
            ]["instance_" + str(i)] = metrics_dict

        with open(attack_json, "w", encoding="utf-8") as f:
            json.dump(json_formatted, f)

        g = GenerateTextReport()
        g.process_attack_target_json(attack_json, target_json)
        g.export_to_file(output_filename)

        clean_up(target_json)
        clean_up(attack_json)
        clean_up(output_filename)

    def test_whitespace_in_filenames(self):
        """Test to make sure whitespace is removed from the output file when creating the report."""
        json_formatted = get_test_report()

        filename = "test.json"
        output_filename = "filename should be changed.txt"

        with open(filename, "w", encoding="utf-8") as f:
            json.dump(json_formatted, f)

        g = GenerateTextReport()
        g.process_attack_target_json(filename)
        g.export_to_file(output_filename)

        assert os.path.exists("filename should be changed.txt") is False
        assert os.path.exists("filename_should_be_changed.txt") is True

        clean_up(filename)
        clean_up(output_filename)
        clean_up("filename_should_be_changed.txt")

    def test_move_files(self):
        """Test the move_files parameter inside export_to_file."""
        filename = "test.json"
        output_filename = "results.txt"
        dummy_model = "dummy_model.txt"

        json_formatted = get_test_report()

        with open(filename, "w", encoding="utf-8") as f:
            json.dump(json_formatted, f)

        with open(dummy_model, "w", encoding="utf-8") as f:
            f.write("dummy_model")

        g = GenerateTextReport()
        g.process_attack_target_json(filename)
        g.export_to_file(output_filename, move_files=True, release_dir="release_dir")

        # Check when no model name has been provided
        assert os.path.exists(os.path.join("release_dir", output_filename)) is True

        g.export_to_file(
            output_filename,
            move_files=True,
            release_dir="release_dir",
            model_filename=dummy_model,
        )

        # Check model file has been copied (NOT moved)
        assert os.path.exists(os.path.join("release_dir", dummy_model)) is True
        assert os.path.exists(dummy_model) is True

        clean_up(dummy_model)

        clean_up(output_filename)

        png_file = "log_roc.png"
        with open(png_file, "w", encoding="utf-8") as f:
            pass

        assert os.path.exists(png_file) is True

        g = GenerateTextReport()
        g.process_attack_target_json(filename)
        g.export_to_file(
            output_filename,
            move_files=True,
            release_dir="release_dir",
            artefacts_dir="training_artefacts",
        )

        assert os.path.exists(png_file) is False
        assert os.path.exists(os.path.join("training_artefacts", png_file)) is True

        clean_up(filename)
        clean_up(output_filename)
        clean_up("release_dir")
        clean_up("training_artefacts")

    def test_complete_runthrough(self):
        """Test the full process_json file end-to-end when valid parameters are passed."""
        json_formatted = get_test_report()
        _ = self.process_json_from_file(json_formatted)


class TestFinalRecommendationModule(unittest.TestCase):
    """Class which tests the FinalRecommendatiionModule inside attack_report_formatter.py."""

    def test_instance_based(self):
        """Test the process_json function when the target model is an instance based model."""
        json_formatted = get_test_report()
        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        immediate_rejection = returned[0]
        # support_rejection = returned[1]
        # support_release = returned[2]

        self.assertEqual(len(immediate_rejection), 0)

        json_formatted["model_name"] = "SVC"
        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        immediate_rejection = returned[0]
        # support_rejection = returned[1]
        # support_release = returned[2]

        self.assertIn("Model is SVM", immediate_rejection)

        json_formatted["model_name"] = "KNeighborsClassifier"
        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        immediate_rejection = returned[0]
        # support_rejection = returned[1]
        # support_release = returned[2]

        self.assertIn("Model is kNN", immediate_rejection)

    def test_min_samples_leaf(self):
        """Test the process_json function when the target model includes decision trees."""

        # test when min_samples_leaf > 5
        json_formatted = get_test_report()

        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        immediate_rejection = returned[0]
        # support_rejection = returned[1]
        # support_release = returned[2]

        self.assertEqual(len(immediate_rejection), 0)

        # test when min_samples_leaf < 5
        json_formatted["model_params"]["min_samples_leaf"] = 2

        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        # immediate_rejection = returned[0]
        support_rejection = returned[1]
        # support_release = returned[2]

        support_rejection = ", ".join(support_rejection)
        self.assertIn("Min samples per leaf", support_rejection)

    def test_statistically_significant(self):
        """Test the statistically significant AUC p-values check in FinalRecommendationModule."""
        json_formatted = get_test_report()
        json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
            "attack_instance_logger"
        ] = {}

        metrics_dict = {
            "P_HIGHER_AUC": 0.001,
            "AUC": 0.8,
            "ACC": 0.7,
            "FDIF01": 0.2,
            "PDIF01": 1.0,
            "FPR": 1.0,
            "TPR": 0.1,
        }

        for i in range(10):
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
                "attack_instance_logger"
            ]["instance_" + str(i)] = metrics_dict

        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        # immediate_rejection = returned[0]
        support_rejection = returned[1]
        # support_release = returned[2]

        support_rejection = ", ".join(support_rejection)

        self.assertIn(">10% AUC are statistically significant", support_rejection)
        self.assertIn("Attack AUC > threshold", support_rejection)

        metrics_dict["AUC"] = 0.5

        for i in range(10):
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
                "attack_instance_logger"
            ]["instance_" + str(i)] = metrics_dict

        f = FinalRecommendationModule(json_formatted)
        f.process_dict()
        returned = f.get_recommendation()

        # immediate_rejection = returned[0]
        # support_rejection = returned[1]
        support_release = returned[2]

        support_release = ", ".join(support_release)

        self.assertIn("Attack AUC <= threshold", support_release)

    def test_print(self):
        """Test the FinalRecommendationModule printing."""
        json_formatted = get_test_report()
        f = FinalRecommendationModule(json_formatted)
        self.assertEqual("Final Recommendation", str(f))


class TestSummariseUnivariateMetricsModule(unittest.TestCase):
    """Class which tests the SummariseUnivariateMetricsModule inside attack_report_formatter.py."""

    def test_univariate_metrics_module(self):
        """Test the SummariseUnivariateMetricsModule."""
        json_formatted = get_test_report()

        auc_value = 0.8
        acc_value = 0.7
        fdif01_value = 0.2

        metrics_dict = {
            "P_HIGHER_AUC": 0.001,
            "AUC": auc_value,
            "ACC": acc_value,
            "FDIF01": fdif01_value,
            "PDIF01": 1.0,
            "FPR": 1.0,
            "TPR": 0.1,
        }

        for i in range(10):
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
                "attack_instance_logger"
            ]["instance_" + str(i)] = metrics_dict

        f = SummariseUnivariateMetricsModule(json_formatted)
        returned = f.process_dict()

        wca_auc = returned["WorstCaseAttack"]["AUC"]
        for k in wca_auc.keys():
            self.assertAlmostEqual(auc_value, wca_auc[k])

        wca_acc = returned["WorstCaseAttack"]["ACC"]
        for k in wca_acc.keys():
            self.assertAlmostEqual(acc_value, wca_acc[k])

        wca_fdif = returned["WorstCaseAttack"]["FDIF01"]
        for k in wca_fdif.keys():
            self.assertAlmostEqual(fdif01_value, wca_fdif[k])

        self.assertEqual("Summary of Univarite Metrics", str(f))

    def test_print(self):
        """Test the SummariseUnivariateMetricsModule printing."""
        json_formatted = get_test_report()
        f = SummariseUnivariateMetricsModule(json_formatted)
        self.assertEqual("Summary of Univarite Metrics", str(f))


class TestSummariseAUCPvalsModule(unittest.TestCase):
    """Class which tests the SummariseAUCPvalsModule inside attack_report_formatter.py."""

    def test_auc_pvals_module(self):
        """Test the SummariseAUCPvalsModule."""
        json_formatted = get_test_report()
        f = SummariseAUCPvalsModule(json_formatted)
        _ = str(f)
        returned = f.process_dict()

        # test the default correction
        self.assertEqual("bh", returned["correction"])
        self.assertEqual(10, returned["n_total"])

        metrics_dict = {"P_HIGHER_AUC": 0.001}

        for i in range(11):
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"][
                "attack_instance_logger"
            ]["instance_" + str(i)] = metrics_dict

        f = SummariseAUCPvalsModule(json_formatted)
        _ = str(f)
        returned = f.process_dict()
        self.assertEqual(11, returned["n_total"])

        f = SummariseAUCPvalsModule(json_formatted, correction="bo")
        returned = f.process_dict()

        self.assertEqual("bo", returned["correction"])

        with pytest.raises(NotImplementedError):
            f = SummariseAUCPvalsModule(json_formatted, correction="xyzabcd")
            _ = f.process_dict()

        _ = json_formatted["WorstCaseAttack"].pop("attack_experiment_logger")
        f = SummariseAUCPvalsModule(json_formatted)

    def test_print(self):
        """Test the SummariseAUCPvalsModule printing."""
        json_formatted = get_test_report()
        f = SummariseAUCPvalsModule(json_formatted)
        self.assertIn("Summary of AUC p-values", str(f))


class TestSummariseFDIFPvalsModule(unittest.TestCase):
    """Class which tests the SummariseFDIFPvalsModule inside attack_report_formatter.py."""

    def test_fdif_pvals_module(self):
        """Test the SummariseFDIFPvalsModule."""
        json_formatted = get_test_report()
        f = SummariseFDIFPvalsModule(json_formatted)
        returned = f.process_dict()

        self.assertEqual("bh", returned["correction"])
        self.assertEqual(10, returned["n_total"])

        returned = f.get_metric_list(
            json_formatted["WorstCaseAttack"]["attack_experiment_logger"]
        )

        self.assertEqual(10, len(returned))

    def test_print(self):
        """Test the SummariseFDIFPvalsModule printing."""
        json_formatted = get_test_report()
        f = SummariseFDIFPvalsModule(json_formatted)
        self.assertIn("Summary of FDIF p-values", str(f))


class TestLogLogROCModule(unittest.TestCase):
    """Class which tests the LogLogROCModule inside attack_report_formatter.py."""

    def test_loglog_roc_module(self):
        """Test the LogLogROCModule."""
        json_formatted = get_test_report()
        f = LogLogROCModule(json_formatted)
        returned = f.process_dict()

        output_file = (
            f"{json_formatted['log_id']}-{json_formatted['metadata']['attack']}.png"
        )
        self.assertIn(output_file, returned)
        assert os.path.exists(output_file) is True

        clean_up(output_file)

        f = LogLogROCModule(json_formatted, output_folder=".")
        returned = f.process_dict()

        output_file = (
            f"{json_formatted['log_id']}-{json_formatted['metadata']['attack']}.png"
        )
        self.assertIn(output_file, returned)
        assert os.path.exists(output_file) is True

        clean_up(output_file)

    def test_loglog_multiple_files(self):
        """Test the LogLogROCModule with multiple tests."""
        out_json = get_test_report()
        out_json_copy = get_test_report()
        out_json_copy["log_id"] = 2048

        out_json.update(out_json_copy)

        f = LogLogROCModule(out_json, output_folder=".")
        returned = f.process_dict()

        output_file_1 = f"{out_json['log_id']}-{out_json['metadata']['attack']}.png"
        output_file_2 = (
            f"{out_json_copy['log_id']}-{out_json_copy['metadata']['attack']}.png"
        )

        self.assertIn(output_file_1, returned)
        self.assertIn(output_file_2, returned)

        clean_up(output_file_1)
        clean_up(output_file_2)

    def test_print(self):
        """Test the LogLogROCModule printing."""
        json_formatted = get_test_report()
        f = LogLogROCModule(json_formatted)
        self.assertEqual("ROC Log Plot", str(f))



"""Jim Smith October 2022
tests to pick up odd cases not otherwise covered
in code in the attacks folder.
"""

from __future__ import annotations

import numpy as np
import pytest
from fpdf import FPDF

from aisdc.attacks import attack, report
from aisdc.attacks.target import Target
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier

BORDER = 0


def test_superclass():
    """Test that the exceptions are raised if the superclass is called in error."""
    model = SafeDecisionTreeClassifier()
    target = Target(model=model)
    my_attack = attack.Attack()
    with pytest.raises(NotImplementedError):
        my_attack.attack(target)
    with pytest.raises(NotImplementedError):
        print(str(my_attack))  # .__str__()


def test_NumpyArrayEncoder():
    """Conversion routine
    from reports.py.
    """

    i32 = np.int32(2)
    i64 = np.int64(2)
    twoDarray = np.zeros((2, 2))
    my_encoder = report.NumpyArrayEncoder()

    retval = my_encoder.default(i32)
    assert isinstance(retval, int)

    retval = my_encoder.default(i64)
    assert isinstance(retval, int)

    retval = my_encoder.default(twoDarray)
    assert isinstance(retval, list)

    with pytest.raises(TypeError):
        retval = my_encoder.default("a string")


def test_line():
    """Code from report.py."""
    pdf = FPDF()
    pdf.add_page()
    report.line(pdf, "foo")
    pdf.close()


def test_dict():
    """Code from report.py."""
    pdf = FPDF()
    pdf.add_page()
    mydict = {"a": "hello", "b": "world"}
    report._write_dict(  # pylint:disable=protected-access
        pdf, mydict, indent=0, border=BORDER
    )
    pdf.close()



"""Test_worst_case_attack.py
Copyright (C) Jim Smith 2022 <james.smith@uwe.ac.uk>.
"""

from __future__ import annotations

import shutil
import unittest

from aisdc.attacks import failfast, worst_case_attack  # pylint: disable = import-error


class TestFailFast(unittest.TestCase):
    """Class which tests the fail fast functionality of the WortCaseAttack module."""

    def test_parse_boolean_argument(self):
        """Test all comparison operators and both options for attack
        being successful and not successful given a metric and
        comparison operator with a threshold value.
        """
        metrics = {}
        metrics["ACC"] = 0.9
        metrics["AUC"] = 0.8
        metrics["P_HIGHER_AUC"] = 0.05

        # Option 1
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.04,
            attack_metric_success_comp_type="lte",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 2
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="lte",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        # Option 3
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.04,
            attack_metric_success_comp_type="lt",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 4
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="lt",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        # Option 5
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.04,
            attack_metric_success_comp_type="gte",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        # Option 6
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="gte",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 7
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.04,
            attack_metric_success_comp_type="gt",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        # Option 8
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="gt",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 9
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.05,
            attack_metric_success_comp_type="eq",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        # Option 10
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="eq",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 11
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.05,
            attack_metric_success_comp_type="not_eq",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertFalse(failfast_Obj.check_attack_success(metrics))

        # Option 12
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.06,
            attack_metric_success_comp_type="not_eq",
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        self.assertTrue(failfast_Obj.check_attack_success(metrics))

        self.assertEqual(0, failfast_Obj.get_fail_count())

        shutil.rmtree("output_worstcase")

    def test_attack_success_fail_counts_and_overall_attack_success(self):
        """Test success and fail counts of attacks for a given threshold
        of a given metric based on a given comparison operation and
        also test overall attack successes using
        count threshold of attack being successful or not successful.
        """
        metrics = {}
        metrics["ACC"] = 0.9
        metrics["AUC"] = 0.8
        metrics["P_HIGHER_AUC"] = 0.08
        attack_obj = worst_case_attack.WorstCaseAttack(
            attack_metric_success_name="P_HIGHER_AUC",
            attack_metric_success_thresh=0.05,
            attack_metric_success_comp_type="lte",
            attack_metric_success_count_thresh=3,
        )
        failfast_Obj = failfast.FailFast(attack_obj)
        _ = failfast_Obj.check_attack_success(metrics)
        metrics["P_HIGHER_AUC"] = 0.07
        _ = failfast_Obj.check_attack_success(metrics)
        metrics["P_HIGHER_AUC"] = 0.03
        _ = failfast_Obj.check_attack_success(metrics)
        self.assertFalse(failfast_Obj.check_overall_attack_success(attack_obj))

        metrics["P_HIGHER_AUC"] = 0.02
        _ = failfast_Obj.check_attack_success(metrics)
        metrics["P_HIGHER_AUC"] = 0.01
        _ = failfast_Obj.check_attack_success(metrics)
        self.assertEqual(3, failfast_Obj.get_success_count())
        self.assertEqual(2, failfast_Obj.get_fail_count())
        self.assertTrue(failfast_Obj.check_overall_attack_success(attack_obj))

        shutil.rmtree("output_worstcase")



"""
Tests attacks called via safemodel classes
uses a subsampled nursery dataset as this tests more of the attack code
currently using random forests.
"""

from __future__ import annotations

import os
import shutil

import numpy as np
import sklearn
from sklearn.datasets import fetch_openml
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder

from aisdc.attacks import attribute_attack, likelihood_attack, worst_case_attack
from aisdc.attacks.target import Target
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier

# pylint: disable=too-many-locals,bare-except,duplicate-code,unnecessary-dunder-call

RES_DIR = "RES"


def clean():
    """Removes unwanted results."""
    if os.path.exists(RES_DIR):
        shutil.rmtree(RES_DIR)


def get_target(model: sklearn.base.BaseEstimator) -> Target:
    """Wrap the model and data in a Target object.
    Uses A randomly sampled 10+10% of the nursery data set.
    """

    nursery_data = fetch_openml(data_id=26, as_frame=True)
    x = np.asarray(nursery_data.data, dtype=str)
    y = np.asarray(nursery_data.target, dtype=str)
    # change labels from recommend to priority for the two odd cases
    num = len(y)
    for i in range(num):
        if y[i] == "recommend":
            y[i] = "priority"

    indices: list[list[int]] = [
        [0, 1, 2],  # parents
        [3, 4, 5, 6, 7],  # has_nurs
        [8, 9, 10, 11],  # form
        [12, 13, 14, 15],  # children
        [16, 17, 18],  # housing
        [19, 20],  # finance
        [21, 22, 23],  # social
        [24, 25, 26],  # health
        [27],  # dummy
    ]

    # [Researcher] Split into training and test sets
    # target model train / test split - these are strings
    (
        x_train_orig,
        x_test_orig,
        y_train_orig,
        y_test_orig,
    ) = train_test_split(
        x,
        y,
        test_size=0.05,
        stratify=y,
        shuffle=True,
    )

    # now resample the training data reduce number of examples
    _, x_train_orig, _, y_train_orig = train_test_split(
        x_train_orig,
        y_train_orig,
        test_size=0.05,
        stratify=y_train_orig,
        shuffle=True,
    )

    # [Researcher] Preprocess dataset
    # one-hot encoding of features and integer encoding of labels
    label_enc = LabelEncoder()
    feature_enc = OneHotEncoder()
    x_train = feature_enc.fit_transform(x_train_orig).toarray()
    y_train = label_enc.fit_transform(y_train_orig)
    x_test = feature_enc.transform(x_test_orig).toarray()
    y_test = label_enc.transform(y_test_orig)

    # add dummy continuous valued attribute from N(0.5,0.05)
    dummy_tr = 0.5 + 0.05 * np.random.randn(x_train.shape[0])
    dummy_te = 0.5 + 0.05 * np.random.randn(x_test.shape[0])
    dummy_all = np.hstack((dummy_tr, dummy_te)).reshape(-1, 1)
    dummy_tr = dummy_tr.reshape(-1, 1)
    dummy_te = dummy_te.reshape(-1, 1)

    x_train = np.hstack((x_train, dummy_tr))
    x_train_orig = np.hstack((x_train_orig, dummy_tr))
    x_test = np.hstack((x_test, dummy_te))
    x_test_orig = np.hstack((x_test_orig, dummy_te))
    xmore = np.concatenate((x_train_orig, x_test_orig))
    n_features = np.shape(x_train_orig)[1]

    # wrap
    target = Target(model=model)
    target.name = "nursery"
    target.add_processed_data(x_train, y_train, x_test, y_test)
    for i in range(n_features - 1):
        target.add_feature(nursery_data.feature_names[i], indices[i], "onehot")
    target.add_feature("dummy", indices[n_features - 1], "float")
    target.add_raw_data(xmore, y, x_train_orig, y_train_orig, x_test_orig, y_test_orig)
    return target


def test_attacks_via_request_release():
    """Make vulnerable,hacked model then call request_release."""
    # build a broken model and hack it so lots of reasons to fail and be vulnerable
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=10, min_samples_leaf=1)
    target = get_target(model)
    assert target.__str__() == "nursery"
    model.fit(target.x_train, target.y_train)
    model.min_samples_leaf = 10
    model.request_release(path=RES_DIR, ext="pkl", target=target)
    clean()


def test_run_attack_lira():
    """Calls the lira attack via safemodel."""
    # build a model
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5, min_samples_leaf=10)
    target = get_target(model)
    model.fit(target.x_train, target.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive

    print(np.unique(target.y_test, return_counts=True))
    print(np.unique(model.predict(target.x_test), return_counts=True))
    metadata = model.run_attack(target, "lira", RES_DIR, "lira_res")
    clean()
    assert len(metadata) > 0  # something has been added


def test_run_attack_worstcase():
    """Calls the worst case attack via safemodel."""
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5, min_samples_leaf=20)
    target = get_target(model)
    model.fit(target.x_train, target.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive
    metadata = model.run_attack(target, "worst_case", RES_DIR, "wc_res")
    clean()
    assert len(metadata) > 0  # something has been added


def test_run_attack_attribute():
    """Calls the attribute  attack via safemodel."""
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5, min_samples_leaf=10)
    target = get_target(model)
    model.fit(target.x_train, target.y_train)
    _, disclosive = model.preliminary_check()
    assert not disclosive
    metadata = model.run_attack(target, "attribute", RES_DIR, "attr_res")
    clean()
    assert len(metadata) > 0  # something has been added


def test_attack_args():
    """Tests the attack arguments class."""
    fname = "aia_example"
    attack_obj = attribute_attack.AttributeAttack(
        output_dir="output_attribute", report_name=fname
    )
    attack_obj.__dict__["foo"] = "boo"
    assert attack_obj.__dict__["foo"] == "boo"
    assert fname == attack_obj.report_name

    fname = "liraa"
    attack_obj = likelihood_attack.LIRAAttack(
        output_dir="output_lira", report_name=fname
    )
    attack_obj.__dict__["foo"] = "boo"
    assert attack_obj.__dict__["foo"] == "boo"
    assert fname == attack_obj.report_name

    fname = "wca"
    attack_obj = worst_case_attack.WorstCaseAttack(
        output_dir="output_worstcase", report_name=fname
    )
    attack_obj.__dict__["foo"] = "boo"
    assert attack_obj.__dict__["foo"] == "boo"
    assert fname == attack_obj.report_name
    shutil.rmtree("output_attribute")
    shutil.rmtree("output_lira")
    shutil.rmtree("output_worstcase")


def test_run_attack_unknown():
    """Calls an unknown attack via safemodel."""
    # build a model
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=5)
    target = get_target(model)
    model.fit(target.x_train, target.y_train)
    metadata = model.run_attack(target, "unknown", RES_DIR, "unk")
    clean()
    assert metadata["outcome"] == "unrecognised attack type requested"



"""Generate report for TREs from JSON file."""

from __future__ import annotations

import json
import os
import pathlib
import pprint
import shutil
from datetime import date

import matplotlib.pyplot as plt
import numpy as np


def cleanup_files_for_release(
    move_into_artefacts,
    copy_into_release,
    release_dir="release_files",
    artefacts_dir="training_artefacts",
):
    """
    Function that will move any files created throughout the release process and
    sort them into appropriate folders.
    """

    if not os.path.exists(release_dir):
        os.makedirs(release_dir)

    if not os.path.exists(artefacts_dir):
        os.makedirs(artefacts_dir)

    for filepath in move_into_artefacts:
        if os.path.exists(str(filepath)):
            filename = os.path.basename(filepath)
            dest = os.path.join(artefacts_dir, filename)
            shutil.move(filename, dest)

    for filepath in copy_into_release:
        if os.path.exists(str(filepath)):
            filename = os.path.basename(filepath)
            dest = os.path.join(release_dir, filename)
            shutil.copy(filepath, dest)


class GenerateJSONModule:
    """Module that creates and appends to a JSON file."""

    def __init__(self, filename=None):
        self.filename = filename

        if self.filename is None:
            self.filename = (
                "ATTACK_RESULTS" + str(date.today().strftime("%d_%m_%Y")) + ".json"
            )

        dirname = os.path.normpath(os.path.dirname(self.filename))
        os.makedirs(dirname, exist_ok=True)
        # if file doesn't exist, create it
        if not os.path.exists(self.filename):
            with open(self.filename, "w", encoding="utf-8") as f:
                f.write("")

    def add_attack_output(self, incoming_json, class_name):
        """Add a section of JSON to the file which is already open."""

        # Read the contents of the file and then clear the file
        with open(self.filename, "r+", encoding="utf-8") as f:
            file_contents = f.read()
            if file_contents != "":
                file_data = json.loads(file_contents)
            else:
                file_data = {}

            f.truncate(0)

        # Add the new JSON to the JSON that was in the file, and re-write
        with open(self.filename, "w", encoding="utf-8") as f:
            incoming_json = json.loads(incoming_json)

            if "log_id" in incoming_json:
                class_name = class_name + "_" + str(incoming_json["log_id"])

            file_data[class_name] = incoming_json
            json.dump(file_data, f)

    def get_output_filename(self):
        """Returns the filename of the JSON file which has been created."""
        return self.filename

    def clean_file(self):
        """Delete the file if it exists."""
        if os.path.exists(self.filename):
            os.remove(self.filename)

        with open(self.filename, "w", encoding="utf-8") as f:
            f.write("")


class AnalysisModule:
    """Wrapper module for metrics analysis modules."""

    def __init__(self):
        self.immediate_rejection = []
        self.support_rejection = []
        self.support_release = []

    def process_dict(self):
        """Function that produces a risk summary output based on analysis in this module."""
        raise NotImplementedError()

    def get_recommendation(self):
        """Function that returns the three recommendation buckets created by this module."""
        return self.immediate_rejection, self.support_rejection, self.support_release

    def __str__(self):
        raise NotImplementedError()


class FinalRecommendationModule(
    AnalysisModule
):  # pylint: disable=too-many-instance-attributes
    """Module that generates the first layer of a recommendation report."""

    def __init__(self, report: dict):
        super().__init__()

        self.P_VAL_THRESH = 0.05
        self.MEAN_AUC_THRESH = 0.65

        self.INSTANCE_MODEL_WEIGHTING_SCORE = 5
        self.MIN_SAMPLES_LEAF_SCORE = 5
        self.STATISTICALLY_SIGNIFICANT_SCORE = 2
        self.MEAN_AUC_SCORE = 4

        self.report = report

        self.scores = []
        self.reasons = []

    def _is_instance_based_model(self, instance_based_model_score):
        if "model_name" in self.report:
            if self.report["model_name"] == "SVC":
                self.scores.append(instance_based_model_score)
                self.reasons.append("Model is SVM (supersedes all other tests)")
                self.immediate_rejection.append("Model is SVM")
                return True
            if self.report["model_name"] == "KNeighborsClassifier":
                self.scores.append(instance_based_model_score)
                self.reasons.append("Model is kNN (supersedes all other tests)")
                self.immediate_rejection.append("Model is kNN")
                return True
        return False

    def _tree_min_samples_leaf(self, min_samples_leaf_score):
        # Find min samples per leaf requirement
        base_path = pathlib.Path(__file__).parents[1]
        risk_appetite_path = os.path.join(base_path, "safemodel", "rules.json")
        min_samples_leaf_appetite = None

        with open(risk_appetite_path, "r+", encoding="utf-8") as f:
            file_contents = f.read()
            json_structure = json.loads(file_contents)

            rules = json_structure["DecisionTreeClassifier"]["rules"]
            for entry in rules:
                if "keyword" in entry.keys() and entry["keyword"] == "min_samples_leaf":
                    if "operator" in entry.keys() and entry["operator"] == "min":
                        min_samples_leaf_appetite = entry["value"]
                        break

        if ("model_params" in self.report) and min_samples_leaf_appetite is not None:
            if "min_samples_leaf" in self.report["model_params"]:
                min_samples_leaf = self.report["model_params"]["min_samples_leaf"]
                if min_samples_leaf < min_samples_leaf_appetite:
                    self.scores.append(min_samples_leaf_score)

                    msg = "Min samples per leaf < " + str(min_samples_leaf_appetite)
                    self.reasons.append(msg)
                    self.support_rejection.append(msg)
                else:
                    msg = "Min samples per leaf > " + str(min_samples_leaf_appetite)
                    self.support_release.append(msg)

    def _statistically_significant_auc(
        self, p_val_thresh, mean_auc_thresh, stat_sig_score, mean_auc_score
    ):
        stat_sig_auc = []
        for k in self.report.keys():
            if isinstance(self.report[k], dict):
                if "attack_experiment_logger" in self.report[k]:
                    for i in self.report[k]["attack_experiment_logger"][
                        "attack_instance_logger"
                    ]:
                        instance = self.report[k]["attack_experiment_logger"][
                            "attack_instance_logger"
                        ][i]

                        auc_key = "P_HIGHER_AUC"
                        if (
                            auc_key in instance.keys()
                            and instance[auc_key] < p_val_thresh
                        ):
                            stat_sig_auc.append(instance["AUC"])

                    n_instances = len(
                        self.report[k]["attack_experiment_logger"][
                            "attack_instance_logger"
                        ]
                    )
                    if (
                        len(stat_sig_auc) / n_instances > 0.1
                    ):  # > 10% of AUC are statistically significant
                        msg = (
                            ">10% AUC are statistically significant in experiment "
                            + str(k)
                        )

                        self.scores.append(stat_sig_score)
                        self.reasons.append(msg)
                        self.support_rejection.append(msg)
                    else:
                        msg = (
                            "<10% AUC are statistically significant in experiment "
                            + str(k)
                        )
                        self.support_release.append(msg)

                    if len(stat_sig_auc) > 0:
                        mean = np.mean(np.array(stat_sig_auc))
                        if mean > mean_auc_thresh:
                            msg = "Attack AUC > threshold of " + str(mean_auc_thresh)
                            msg = msg + " in experiment " + str(k)

                            self.scores.append(mean_auc_score)
                            self.reasons.append(msg)
                            self.support_rejection.append(msg)
                        else:
                            msg = "Attack AUC <= threshold of " + str(mean_auc_thresh)
                            msg = msg + " in experiment " + str(k)
                            self.support_release.append(msg)

    def process_dict(self):
        self._tree_min_samples_leaf(self.MIN_SAMPLES_LEAF_SCORE)
        self._statistically_significant_auc(
            self.P_VAL_THRESH,
            self.MEAN_AUC_THRESH,
            self.STATISTICALLY_SIGNIFICANT_SCORE,
            self.MEAN_AUC_SCORE,
        )

        if len(self.scores) == 0:
            summarised_score = 0
        else:
            summarised_score = int(np.sum(np.array(self.scores)).round(0))
            summarised_score = min(summarised_score, 5)

        # if model is instance based, it is automatically disclosive. Assign max score
        if self._is_instance_based_model(self.INSTANCE_MODEL_WEIGHTING_SCORE):
            summarised_score = self.INSTANCE_MODEL_WEIGHTING_SCORE

        output = {}

        msg = "Final score (scale of 0-5, where 0 is least disclosive and 5 is recommend rejection)"
        output[msg] = summarised_score

        return output

    def __str__(self):
        return "Final Recommendation"


class SummariseUnivariateMetricsModule(AnalysisModule):
    """Module that summarises a set of chosen univariate metrics from the output dictionary."""

    def __init__(self, report: dict, metrics_list=None):
        super().__init__()

        if metrics_list is None:
            metrics_list = ["AUC", "ACC", "FDIF01"]

        self.report = report
        self.metrics_list = metrics_list

    def process_dict(self):
        output_dict = {}

        for k in self.report.keys():
            if isinstance(self.report[k], dict):
                if "attack_experiment_logger" in self.report[k]:
                    metrics_dict = {m: [] for m in self.metrics_list}
                    for _, iteration_value in self.report[k][
                        "attack_experiment_logger"
                    ]["attack_instance_logger"].items():
                        for m in metrics_dict:
                            metrics_dict[m].append(iteration_value[m])
                    output = {}
                    for m in self.metrics_list:
                        output[m] = {
                            "min": min(metrics_dict[m]),
                            "max": max(metrics_dict[m]),
                            "mean": np.mean(metrics_dict[m]),
                            "median": np.median(metrics_dict[m]),
                        }
                    output_dict[k] = output
        return output_dict

    def __str__(self):
        return "Summary of Univarite Metrics"


class SummariseAUCPvalsModule(AnalysisModule):
    """Module that summarises a list of AUC values."""

    def __init__(self, report: dict, p_thresh: float = 0.05, correction: str = "bh"):
        super().__init__()

        self.report = report
        self.p_thresh = p_thresh
        self.correction = correction

    def _n_sig(self, p_val_list: list[float], correction: str = "none") -> int:
        """Compute the number of significant p-vals in a list with different corrections for
        multiple testing.
        """

        if correction == "none":
            return len(np.where(np.array(p_val_list) <= self.p_thresh)[0])
        if correction == "bh":
            sorted_p = sorted(p_val_list)
            m = len(p_val_list)
            alpha = self.p_thresh
            comparators = np.arange(1, m + 1, 1) * alpha / m
            return (sorted_p <= comparators).sum()
        if correction == "bo":  # bonferroni
            return len(
                np.where(np.array(p_val_list) <= self.p_thresh / len(p_val_list))[0]
            )

        raise NotImplementedError()  # any others?

    def _get_metrics_list(self) -> list[float]:
        metrics_list = []
        for k in self.report.keys():
            if isinstance(self.report[k], dict):
                if "attack_experiment_logger" in self.report[k]:
                    for _, iteration_value in self.report[k][
                        "attack_experiment_logger"
                    ]["attack_instance_logger"].items():
                        metrics_list.append(iteration_value["P_HIGHER_AUC"])
        return metrics_list

    def process_dict(self):
        """Process the dict to summarise the number of significant AUC p-values."""
        p_val_list = self._get_metrics_list()
        output = {
            "n_total": len(p_val_list),
            "p_thresh": self.p_thresh,
            "n_sig_uncorrected": self._n_sig(p_val_list),
            "correction": self.correction,
            "n_sig_corrected": self._n_sig(p_val_list, self.correction),
        }
        return output

    def __str__(self):
        return f"Summary of AUC p-values at p = ({self.p_thresh})"


class SummariseFDIFPvalsModule(SummariseAUCPvalsModule):
    """Summarise the number of significant FDIF p-values."""

    def get_metric_list(self, input_dict: dict) -> list[float]:
        """Get metrics_list from attack_instance_logger within JSON file."""
        metric_list = []
        for _, iteration_value in input_dict["attack_instance_logger"].items():
            metric_list.append(iteration_value["PDIF01"])
        metric_list = [np.exp(-m) for m in metric_list]
        return metric_list

    def __str__(self):
        return f"Summary of FDIF p-values at p = ({self.p_thresh})"


class LogLogROCModule(AnalysisModule):
    """Module that generates a log-log plot."""

    def __init__(self, report: dict, output_folder=None, include_mean=True):
        super().__init__()

        self.report = report
        self.output_folder = output_folder
        self.include_mean = include_mean

    def process_dict(self):
        """Create a roc plot for multiple repetitions."""
        log_plot_names = []

        for k in self.report.keys():
            if isinstance(self.report[k], dict):
                if "attack_experiment_logger" in self.report[k]:
                    plt.figure(figsize=(8, 8))
                    plt.plot([0, 1], [0, 1], "k--")

                    # Compute average ROC
                    base_fpr = np.linspace(0, 1, 1000)
                    metrics = self.report[k]["attack_experiment_logger"][
                        "attack_instance_logger"
                    ].values()
                    all_tpr = np.zeros((len(metrics), len(base_fpr)), float)

                    for i, metric_set in enumerate(metrics):
                        all_tpr[i, :] = np.interp(
                            base_fpr, metric_set["fpr"], metric_set["tpr"]
                        )

                    for _, metric_set in enumerate(metrics):
                        plt.plot(
                            metric_set["fpr"],
                            metric_set["tpr"],
                            color="lightsalmon",
                            linewidth=0.5,
                        )

                    tpr_mu = all_tpr.mean(axis=0)
                    plt.plot(base_fpr, tpr_mu, "r")

                    plt.xlabel("False Positive Rate")
                    plt.ylabel("True Positive Rate")
                    plt.xscale("log")
                    plt.yscale("log")
                    plt.tight_layout()
                    plt.grid()
                    out_file = f"{self.report['log_id']}-{self.report['metadata']['attack']}.png"
                    if self.output_folder is not None:
                        out_file = os.path.join(self.output_folder, out_file)
                    plt.savefig(out_file)
                    log_plot_names.append(out_file)
        msg = "Log plot(s) saved to " + str(log_plot_names)
        return msg

    def __str__(self):
        return "ROC Log Plot"


class GenerateTextReport:
    """Module that generates a text report from a JSON input."""

    def __init__(self):
        self.text_out = []
        self.target_json_filename = None
        self.attack_json_filename = None
        self.model_name_from_target = None

        self.immediate_rejection = []
        self.support_rejection = []
        self.support_release = []

    def _pretty_print(self, report: dict, title) -> str:
        """Function that formats JSON code to make it more readable for TREs."""

        returned_string = str(title) + "\n"

        for key in report.keys():
            returned_string = returned_string + key + "\n"
            returned_string = returned_string + pprint.pformat(report[key]) + "\n\n"

        return returned_string

    def _process_target_json(self):
        """Function that creates a summary of a target model JSON file."""

        model_params_of_interest = [
            "C",
            "kernel",
            "n_neighbors",
            "hidden_layer_sizes",
            "activation",
            "max_depth",
            "min_samples_split",
            "min_samples_leaf",
            "n_estimators",
            "learning_rate",
        ]

        with open(self.target_json_filename, encoding="utf-8") as f:
            json_report = json.loads(f.read())

        output_string = "TARGET MODEL SUMMARY\n"

        if "model_name" in json_report.keys():
            output_string = (
                output_string + "model_name: " + json_report["model_name"] + "\n"
            )

        if "n_samples" in json_report.keys():
            output_string = output_string + "number of samples used to train: "
            output_string = output_string + str(json_report["n_samples"]) + "\n"

        if "model_params" in json_report.keys():
            for param in model_params_of_interest:
                if param in json_report["model_params"].keys():
                    output_string = output_string + param + ": "
                    output_string = output_string + str(
                        json_report["model_params"][param]
                    )
                    output_string = output_string + "\n"

        if "model_path" in json_report.keys():
            filepath = os.path.split(os.path.abspath(self.target_json_filename))[0]
            self.model_name_from_target = os.path.join(
                filepath, json_report["model_path"]
            )

        self.text_out.append(output_string)

    def process_attack_target_json(
        self, attack_filename: str, target_filename: str = None
    ):
        """Function that creates a neat summary of an attack JSON file."""
        self.attack_json_filename = attack_filename

        with open(attack_filename, encoding="utf-8") as f:
            json_report = json.loads(f.read())

        if target_filename is not None:
            self.target_json_filename = target_filename

            with open(target_filename, encoding="utf-8") as f:
                target_file = json.loads(f.read())
                json_report = {**json_report, **target_file}

        modules = [
            FinalRecommendationModule(json_report),
        ]

        for m in modules:
            output = m.process_dict()
            returned = m.get_recommendation()

            self.immediate_rejection += returned[0]
            self.support_rejection += returned[1]
            self.support_release += returned[2]

        output_string = self._pretty_print(output, "ATTACK JSON RESULTS")

        self.text_out.append(output_string)

        bucket_text = "Immediate rejection recommended for the following reason:\n"
        if len(self.immediate_rejection) > 0:
            for reason in self.immediate_rejection:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        bucket_text += "\nEvidence supporting rejection:\n"
        if len(self.support_rejection) > 0:
            for reason in self.support_rejection:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        bucket_text += "\nEvidence supporting release:\n"
        if len(self.support_release) > 0:
            for reason in self.support_release:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        self.text_out.append(bucket_text)

    def export_to_file(  # pylint: disable=too-many-arguments
        self,
        output_filename: str = "summary.txt",
        move_files=False,
        model_filename=None,
        release_dir="release_files",
        artefacts_dir="training_artefacts",
    ):
        """Function that takes the input strings collected and combines into a neat text file."""

        copy_of_text_out = self.text_out
        self.text_out = []

        if self.target_json_filename is not None:
            self._process_target_json()

        self.text_out += copy_of_text_out

        output_filename = output_filename.replace(" ", "_")

        with open(output_filename, "w", encoding="utf-8") as text_file:
            for output_string in self.text_out:
                text_file.write(output_string)
                text_file.write("\n")

        if move_files is True:
            move_into_artefacts = ["log_roc.png"]

            copy_into_release = [
                output_filename,
                self.attack_json_filename,
                self.target_json_filename,
            ]

            if model_filename is None:
                copy_into_release.append(self.model_name_from_target)
            else:
                copy_into_release.append(model_filename)

            cleanup_files_for_release(
                move_into_artefacts, copy_into_release, release_dir, artefacts_dir
            )



"""
Structural_attack.py.

Runs a number of 'static' structural attacks,based on:
- the target model's properties
- the TREs risk appetite as applied to tables and standard regressions
"""

from __future__ import annotations

import argparse
import logging
import os
import uuid
from datetime import datetime

import numpy as np
from acro import ACRO

# tree-based model types currently supported
from sklearn.base import BaseEstimator
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost.sklearn import XGBClassifier

from aisdc.attacks import report
from aisdc.attacks.attack import Attack
from aisdc.attacks.attack_report_formatter import GenerateJSONModule
from aisdc.attacks.target import Target

logging.basicConfig(level=logging.INFO)


def get_unnecessary_risk(model: BaseEstimator) -> bool:
    """
    Checks whether a model's hyper-parameters against
    a set of rules that predict the top 20% most risky.

    This check is designed to assess whether a model is
    likely to be **unnecessarily** risky, i.e.,
    whether it is highly likely that a different combination of hyper-parameters
    would have led to model with similar or better accuracy on the task
    but with lower membership inference risk.

    The rules applied from an experimental study using a grid search in which:
    - max_features was one-hot encoded from the set [None, log2, sqrt]
    - splitter was encoded using 0=best, 1=random

    The target models created were then subject to membership inference attacks (MIA)
    and the hyper-param combinations rank-ordered according to MIA AUC.
    Then a decision tree trained to recognise whether
    hyper-params combintions were in the 20% most risky.
    The rules below were extracted from that tree for the 'least risky' nodes
    """
    # Returns 1 if high risk, otherwise 0
    if not isinstance(
        model, (DecisionTreeClassifier, RandomForestClassifier, XGBClassifier)
    ):
        return 0  # no experimental evidence to support rejection
    unnecessary_risk = 0
    max_depth = float(model.max_depth) if model.max_depth else 500

    # pylint:disable=chained-comparison,too-many-boolean-expressions
    if isinstance(model, DecisionTreeClassifier):
        max_features = model.max_features
        min_samples_leaf = model.min_samples_leaf
        min_samples_split = model.min_samples_split
        splitter = model.splitter
        if (
            (max_depth > 7.5 and min_samples_leaf <= 7.5 and min_samples_split <= 15)
            or (
                splitter == "best"
                and max_depth > 7.5
                and min_samples_leaf <= 7.5
                and min_samples_split > 15
            )
            or (
                splitter == "best"
                and max_depth > 7.5
                and 7.5 < min_samples_leaf <= 15
                and max_features is None
            )
            or (
                splitter == "best"
                and 3.5 < max_depth <= 7.5
                and max_features is None
                and min_samples_leaf <= 7.5
            )
            or (
                splitter == "random"
                and max_depth > 7.5
                and min_samples_leaf <= 7.5
                and max_features is None
            )
        ):
            unnecessary_risk = 1
    elif isinstance(model, RandomForestClassifier):
        n_estimators = model.n_estimators
        max_features = model.max_features
        min_samples_leaf = model.min_samples_leaf
        min_samples_split = model.min_samples_split
        if (
            (max_depth > 3.5 and n_estimators > 35 and max_features is not None)
            or (
                max_depth > 3.5
                and n_estimators > 35
                and min_samples_split <= 15
                and max_features is None
                and model.bootstrap
            )
            or (
                max_depth > 7.5
                and 15 < n_estimators <= 35
                and min_samples_leaf <= 15
                and not model.bootstrap
            )
        ):
            unnecessary_risk = 1

    elif isinstance(model, XGBClassifier):
        n_estimators = model.n_estimators
        if (
            (
                max_depth > 3.5
                and 3.5 < n_estimators <= 12.5
                and model.min_child_weight <= 1.5
            )
            or (max_depth > 3.5 and n_estimators > 12.5 and model.min_child_weight <= 3)
            or (
                max_depth > 3.5
                and n_estimators > 62.5
                and 3 < model.min_child_weight <= 6
            )
        ):
            unnecessary_risk = 1
    return unnecessary_risk


def get_tree_parameter_count(dtree: DecisionTreeClassifier) -> int:
    """Reads the tree structure a returns the number of learned parameters."""
    n_nodes = dtree.tree_.node_count
    left = dtree.tree_.children_left
    right = dtree.tree_.children_right
    is_leaf = np.zeros(n_nodes, dtype=int)
    for node_id in range(n_nodes):
        if left[node_id] == right[node_id]:
            is_leaf[node_id] = 1
    n_leaves = is_leaf.sum()

    # degrees of freedom
    n_internal_nodes = n_nodes - n_leaves
    n_params = 2 * n_internal_nodes  # feature id and threshold
    n_params += n_leaves * (dtree.n_classes_ - 1)  # probability distribution
    return n_params


def get_model_param_count(model: BaseEstimator) -> int:
    """Returns the number of trained parameters in a model."""
    n_params = 0

    if isinstance(model, DecisionTreeClassifier):
        n_params = get_tree_parameter_count(model)

    elif isinstance(model, RandomForestClassifier):
        for member in model.estimators_:
            n_params += get_tree_parameter_count(member)

    elif isinstance(model, AdaBoostClassifier):
        try:  # sklearn v1.2+
            base = model.estimator
        except AttributeError:  # sklearn version <1.2
            base = model.base_estimator
        if isinstance(base, DecisionTreeClassifier):
            for member in model.estimators_:
                n_params += get_tree_parameter_count(member)

    # TO-DO define these for xgb, logistic regression, SVC and others
    elif isinstance(model, XGBClassifier):
        df = model.get_booster().trees_to_dataframe()
        n_trees = df["Tree"].max()
        total = len(df)
        n_leaves = len(df[df.Feature == "Leaf"])
        # 2 per internal node, one per clas in leaves, one weight per tree
        n_params = 2 * (total - n_leaves) + (model.n_classes_ - 1) * n_leaves + n_trees
    else:
        pass

    return n_params


class StructuralAttack(Attack):
    """Class to wrap a number of attacks based on the static structure of a model."""

    # pylint: disable=too-many-instance-attributes

    def __init__(  # pylint: disable = too-many-arguments, too-many-locals
        self,
        attack_config_json_file_name: str = None,
        risk_appetite_config: str = "default",
        target_path: str = None,
        output_dir="outputs_structural",
        report_name="report_structural",
    ) -> None:
        """Constructs an object to execute a structural attack.

        Parameters
        ----------
        report_name : str
            name of the pdf and json output reports
        target_path : str
            path to the saved trained target model and target data
        risk_appetite_config : str
            path to yaml file specifying TRE risk appetite
        """

        super().__init__()
        logger = logging.getLogger("structural_attack")
        self.target: Target = None
        self.target_path = target_path
        self.attack_config_json_file_name = attack_config_json_file_name
        # disclosure risk
        self.k_anonymity_risk = 0
        self.DoF_risk = 0
        self.unnecessary_risk = 0
        self.class_disclosure_risk = 0
        self.lowvals_cd_risk = 0
        self.metadata = {}
        # make dummy acro object and use it to extract risk appetite
        myacro = ACRO(risk_appetite_config)
        self.risk_appetite_config = risk_appetite_config
        self.THRESHOLD = myacro.config["safe_threshold"]
        self.DOF_THRESHOLD = myacro.config["safe_dof_threshold"]
        logger.info(
            "Thresholds for count %i and Dof %i", self.THRESHOLD, self.DOF_THRESHOLD
        )
        del myacro
        if self.attack_config_json_file_name is not None:
            self._update_params_from_config_file()

        # metrics
        self.attack_metrics = [
            "DoF_risk",
            "k_anonymity_risk",
            "class_disclosure_risk",
            "lowvals_cd_risk",
            "unnecessary_risk",
        ]
        self.yprobs = []

        # paths for reporting
        self.output_dir = output_dir
        self.report_name = report_name

    def __str__(self):
        return "Structural attack"

    def attack(self, target: Target) -> None:
        """Programmatic attack entry point.

        To be used when code has access to Target class and trained target model

        Parameters
        ----------
        target : attacks.target.Target
            target as a Target class object
        """
        self.target = target
        if target.model is None:
            errstr = (
                "cannot currently call StructuralAttack.attack() "
                "unless the target contains a trained model"
            )
            raise NotImplementedError(errstr)

        # get proba values for training data
        x = self.target.x_train
        y = self.target.y_train
        assert x.shape[0] == len(y), "length mismatch between trainx and trainy"
        self.yprobs = self.target.model.predict_proba(x)

        # only equivalence classes and membership once as potentially slow
        if isinstance(target.model, DecisionTreeClassifier):
            equiv = self.dt_get_equivalence_classes()
        else:
            equiv = self.get_equivalence_classes()
        equiv_classes = equiv[0]
        equiv_counts = equiv[1]
        equiv_members = equiv[2]
        errstr = "len mismatch between equiv classes and "
        assert len(equiv_classes) == len(equiv_counts), errstr + "counts"
        assert len(equiv_classes) == len(equiv_members), errstr + "membership"
        print(
            f"equiv_classes is {equiv_classes}\n"
            f"equiv_counts is {equiv_counts}\n"
            #   #f'equiv_members is {equiv_members}\n'
        )

        # now assess the risk
        # Degrees of Freedom
        n_params = get_model_param_count(target.model)
        residual_dof = self.target.x_train.shape[0] - n_params
        self.DoF_risk = 1 if residual_dof < self.DOF_THRESHOLD else 0

        # k-anonymity
        mink = np.min(np.array(equiv_counts))
        self.k_anonymity_risk = 1 if mink < self.THRESHOLD else 0

        # unnecessary risk arising from poor hyper-parameter combination.
        self.unnecessary_risk = get_unnecessary_risk(self.target.model)

        # class disclosure
        freqs = np.zeros(equiv_classes.shape)
        for group in range(freqs.shape[0]):
            freqs = equiv_classes[group] * equiv_counts[group]
        self.class_disclosure_risk = np.any(freqs < self.THRESHOLD).astype(int)
        freqs[freqs == 0] = 100
        self.lowvals_cd_risk = np.any(freqs < self.THRESHOLD).astype(int)

    def dt_get_equivalence_classes(self) -> tuple:
        """
        Gets details of equivalence classes
        based on white box inspection.
        """
        destinations = self.target.model.apply(self.target.x_train)
        ret_tuple = np.unique(destinations, return_counts=True)
        # print(f'leaves and counts:\n{ret_tuple}\n')
        leaves = ret_tuple[0]
        counts = ret_tuple[1]
        members = []
        for leaf in leaves:
            ingroup = np.asarray(destinations == leaf).nonzero()[0]
            # print(f'ingroup {ingroup},count {len(ingroup)}')
            members.append(ingroup)

        equiv_classes = np.zeros((len(leaves), self.target.model.n_classes_))
        for group in range(len(leaves)):
            sample_id = members[group][0]
            sample = self.target.x_train[sample_id]
            proba = self.target.model.predict_proba(sample.reshape(1, -1))
            equiv_classes[group] = proba
        return [equiv_classes, counts, members]

    def get_equivalence_classes(self) -> tuple:
        """
        Gets details of equivalence classes
        based on black box observation of probabilities.
        """
        uniques = np.unique(self.yprobs, axis=0, return_counts=True)
        equiv_classes = uniques[0]
        equiv_counts = uniques[1]
        members = []
        for prob_vals in equiv_classes:
            ingroup = np.unique(np.asarray(self.yprobs == prob_vals).nonzero()[0])
            members.append(ingroup)
        return [equiv_classes, equiv_counts, members]

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Summarise metrics from a metric list.

        Returns
        -------
        global_metrics : Dict
            Dictionary of summary metrics

        Arguments
        ---------
        attack_metrics: List
            list of attack metrics to be reported
        """
        global_metrics = {}
        if attack_metrics is not None and len(attack_metrics) != 0:
            global_metrics["DoF_risk"] = self.DoF_risk
            global_metrics["k_anonymity_risk"] = self.k_anonymity_risk
            global_metrics["class_disclosure_risk"] = self.class_disclosure_risk
            global_metrics["unnecessary_risk"] = self.unnecessary_risk
            global_metrics["lowvals_cd_risk"] = self.lowvals_cd_risk

        return global_metrics

    def _construct_metadata(self):
        """Constructs the metadata object, after attacks."""
        self.metadata = {}
        # Store all args
        self.metadata["experiment_details"] = {}
        self.metadata["experiment_details"] = self.get_params()
        self.metadata["attack"] = str(self)
        # Global metrics
        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)

    def _get_attack_metrics_instances(self) -> dict:
        """Constructs the metadata object, after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}

        # for rep, name in enumerate(self.attack_metrics):
        #    #attack_metrics_instances["instance_" + str(rep)] = self.attack_metrics[rep]
        #    attack_metrics_instances["instance_" + str(name)] = self.__dict__.[name]

        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        attack_metrics_experiment["DoF_risk"] = self.DoF_risk
        attack_metrics_experiment["k_anonymity_risk"] = self.k_anonymity_risk
        attack_metrics_experiment["class_disclosure_risk"] = self.class_disclosure_risk
        attack_metrics_experiment["unnecessary_risk"] = self.unnecessary_risk
        attack_metrics_experiment["lowvals_cd_risk"] = self.lowvals_cd_risk
        return attack_metrics_experiment

    def make_report(self) -> dict:
        """Creates output dictionary structure and generates
        pdf and json outputs if filenames are given.
        """
        output = {}
        output["log_id"] = str(uuid.uuid4())
        output["log_time"] = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

        self._construct_metadata()
        output["metadata"] = self.metadata

        output["attack_experiment_logger"] = self._get_attack_metrics_instances()
        # output[
        #     "dummy_attack_experiments_logger"
        # ] = self._get_dummy_attack_metrics_experiments_instances()

        report_dest = os.path.join(self.output_dir, self.report_name)
        json_attack_formatter = GenerateJSONModule(report_dest + ".json")
        json_report = report.create_json_report(output)
        json_attack_formatter.add_attack_output(json_report, "StructuralAttack")

        # pdf_report = report.create_mia_report(output)
        # report.add_output_to_pdf(report_dest, pdf_report, "StructuralAttack")
        return output


def _run_attack(args):
    """Initialise class and run attack."""

    attack_obj = StructuralAttack(
        risk_appetite_config=args.risk_appetite_config,
        target_path=args.target_path,
        output_dir=args.output_dir,
        report_name=args.report_name,
    )

    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    _ = attack_obj.make_report()


def _run_attack_from_configfile(args):
    """Initialise class and run attack  using config file."""

    attack_obj = StructuralAttack(
        attack_config_json_file_name=str(args.attack_config_json_file_name),
        target_path=str(args.target_path),
    )
    target = Target()
    target.load(attack_obj.target_path)
    attack_obj.attack(target)
    _ = attack_obj.make_report()


def main():
    """Main method to parse arguments and invoke relevant method."""
    logger = logging.getLogger("main")
    parser = argparse.ArgumentParser(description="Perform a structural  attack")

    subparsers = parser.add_subparsers()

    attack_parser = subparsers.add_parser("run-attack")

    attack_parser.add_argument(
        "--output-dir",
        type=str,
        action="store",
        dest="output_dir",
        default="output_structural",
        required=False,
        help=("Directory name where output files are stored. Default = %(default)s."),
    )

    attack_parser.add_argument(
        "--report-name",
        type=str,
        action="store",
        dest="report_name",
        default="report_structural",
        required=False,
        help=(
            """Filename for the pdf and json report outputs. Default = %(default)s.
            Code will append .pdf and .json"""
        ),
    )

    attack_parser.add_argument(
        "--risk-appetite-filename",
        action="store",
        type=str,
        default="default",
        required=False,
        dest="risk_appetite_config",
        help=(
            """provide the name of the dataset-specific risk appetite filename
            using --risk-appetite-filename Default = %(default)s"""
        ),
    )

    attack_parser.add_argument(
        "--target-path",
        action="store",
        type=str,
        default=None,
        required=False,
        dest="target_path",
        help=(
            """Provide the path to the stored target usinmg
             --target-path option. Default = %(default)f"""
        ),
    )

    attack_parser.set_defaults(func=_run_attack)

    attack_parser_config = subparsers.add_parser("run-attack-from-configfile")
    attack_parser_config.add_argument(
        "-j",
        "--attack-config-json-file-name",
        action="store",
        required=True,
        dest="attack_config_json_file_name",
        type=str,
        default="config_structural_cmd.json",
        help=(
            "Name of the .json file containing details for the run. Default = %(default)s"
        ),
    )

    attack_parser_config.add_argument(
        "-t",
        "--attack-target-folder-path",
        action="store",
        required=True,
        dest="target_path",
        type=str,
        default="structural_target",
        help=(
            """Name of the target directory to load the trained target model and the target data.
            Default = %(default)s"""
        ),
    )

    attack_parser_config.set_defaults(func=_run_attack_from_configfile)

    args = parser.parse_args()

    try:
        args.func(args)
    except AttributeError as e:  # pragma:no cover
        logger.error("Invalid command. Try --help to get more details")
        logger.error(e)


if __name__ == "__main__":  # pragma:no cover
    main()



"""Failfast.py - class to evaluate metric for fail fast option."""

from __future__ import annotations

from typing import Any


class FailFast:  # pylint: disable=too-many-instance-attributes
    """Class to check attack being successful or not for a given metric
    Note: An object of a FailFast is stateful and instance members
    (success_count and fail_count) will preserve values
    across repetitions for a test. For the new test
    a new object will require to be instantiated.
    """

    def __init__(self, attack_obj: Any):
        self.metric_name = attack_obj.attack_metric_success_name
        self.metric_success_thresh = attack_obj.attack_metric_success_thresh
        self.comp_type = attack_obj.attack_metric_success_comp_type
        self.success_count = 0
        self.fail_count = 0

    # pylint: disable=too-many-branches
    def check_attack_success(self, metric_dict: dict) -> bool:
        """A function to check if attack was successful for a given metric.

        Parameters
        ----------
        metric_dict : dict
            a dictionary with all computed metric values

        Returns
        -------
        success_status : bool
            a boolean value is returned based on the comparison for a given threshold

        Notes
        -----
        If value of a given metric value has a value meeting the threshold based on
        the comparison type returns true otherwise it returns false. This function
        also counts how many times the attack was successful (i.e. true) and
        how many times it was not successful (i.e. false).
        """
        metric_value = metric_dict[self.metric_name]
        success_status = False
        if self.comp_type == "lt":
            success_status = bool(metric_value < self.metric_success_thresh)
        elif self.comp_type == "lte":
            success_status = bool(metric_value <= self.metric_success_thresh)
        elif self.comp_type == "gt":
            success_status = bool(metric_value > self.metric_success_thresh)
        elif self.comp_type == "gte":
            success_status = bool(metric_value >= self.metric_success_thresh)
        elif self.comp_type == "eq":
            success_status = bool(metric_value == self.metric_success_thresh)
        elif self.comp_type == "not_eq":
            success_status = bool(metric_value != self.metric_success_thresh)

        if success_status:
            self._increment_success_count()
        else:
            self._incremenet_fail_count()

        return success_status

    def _increment_success_count(self) -> int:
        self.success_count += 1

    def _incremenet_fail_count(self) -> int:
        self.fail_count += 1

    def get_success_count(self) -> int:
        """Returns a count of attack being successful."""
        return self.success_count

    def get_fail_count(self):
        """Returns a count of attack being not successful."""
        return self.fail_count

    def get_attack_summary(self) -> dict:
        """Returns a dictionary of counts of attack being successful and not successful."""
        summary = {}
        summary["success_count"] = self.success_count
        summary["fail_count"] = self.fail_count
        return summary

    def check_overall_attack_success(self, attack_obj: Any) -> bool:
        """Returns true if the attack is successful for a given success count threshold."""
        overall_success_status = False
        if self.success_count >= attack_obj.attack_metric_success_count_thresh:
            overall_success_status = True
        return overall_success_status



"""Test_worst_case_attack.py
Copyright (C) Jim Smith 2023 <james.smith@uwe.ac.uk>.
"""

from __future__ import annotations

import json
import os
import shutil
import sys
from unittest.mock import patch

# import numpy as np
import pytest
from sklearn.datasets import load_breast_cancer
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from xgboost.sklearn import XGBClassifier

import aisdc.attacks.structural_attack as sa
from aisdc.attacks.target import Target


def clean_up(name):
    """Removes unwanted files or directory."""
    if os.path.exists(name):
        if os.path.isfile(name):
            os.remove(name)
        elif os.path.isdir(name):
            shutil.rmtree(name)


def get_target(modeltype: str, **kwparams: dict) -> Target:
    """Loads dataset and creates target of the desired type."""

    X, y = load_breast_cancer(return_X_y=True, as_frame=False)
    train_X, test_X, train_y, test_y = train_test_split(X, y, test_size=0.3)

    # these types should be handled
    if modeltype == "dt":
        target_model = DecisionTreeClassifier(**kwparams)
    elif modeltype == "rf":
        target_model = RandomForestClassifier(**kwparams)
    elif modeltype == "xgb":
        target_model = XGBClassifier(**kwparams)
    elif modeltype == "adaboost":
        target_model = AdaBoostClassifier(**kwparams)
    # should get polite error but not DoF yet
    elif modeltype == "svc":
        target_model = SVC(**kwparams)
    else:
        raise NotImplementedError("model type passed to get_model unknown")

    # Train the classifier
    target_model.fit(train_X, train_y)

    #  Wrap the model and data in a Target object
    target = Target(model=target_model)
    target.add_processed_data(train_X, train_y, test_X, test_y)

    return target


def test_unnecessary_risk():
    """Checking the unnecessary rules."""
    # non-tree we have no evidence yet
    model = SVC()
    assert sa.get_unnecessary_risk(model) == 0, "no risk without evidence"
    # decision tree next
    risky_param_dicts = [
        {
            "max_features": "log2",
            "max_depth": 8,
            "min_samples_leaf": 7,
            "min_samples_split": 14,
        },
        {
            "splitter": "best",
            "max_depth": 8,
            "min_samples_leaf": 7,
            "min_samples_split": 16,
        },
        {
            "splitter": "best",
            "max_depth": 8,
            "min_samples_leaf": 10,
            "max_features": None,
        },
        {
            "splitter": "best",
            "max_depth": 4,
            "max_features": None,
            "min_samples_leaf": 7,
        },
        {
            "splitter": "random",
            "max_depth": 8,
            "min_samples_leaf": 7,
            "max_features": None,
            "min_samples_split": 25,
        },
    ]
    for idx, paramdict in enumerate(risky_param_dicts):
        model = DecisionTreeClassifier(**paramdict)
        errstr = f" unnecessary risk with rule {idx}" f"params are {model.get_params()}"
        assert sa.get_unnecessary_risk(model) == 1, errstr
    model = DecisionTreeClassifier(max_depth=1, min_samples_leaf=150)
    assert (
        sa.get_unnecessary_risk(model) == 0
    ), f"should be non-disclosive with {model.get_params}"

    # now random forest
    risky_param_dicts = [
        {"max_depth": 50, "n_estimators": 50, "max_features": None},
        {
            "max_depth": 50,
            "n_estimators": 50,
            "min_samples_split": 5,
            "max_features": None,
            "bootstrap": True,
        },
        {
            "max_depth": 25,
            "n_estimators": 25,
            "min_samples_leaf": 5,
            "bootstrap": False,
        },
    ]
    for idx, paramdict in enumerate(risky_param_dicts):
        model = RandomForestClassifier(**paramdict)
        errstr = f" unnecessary risk with rule {idx}" f"params are {model.get_params()}"
        assert sa.get_unnecessary_risk(model) == 1, errstr
    model = RandomForestClassifier(max_depth=1, n_estimators=25, min_samples_leaf=150)
    assert (
        sa.get_unnecessary_risk(model) == 0
    ), f"should be non-disclosive with {model.get_params}"

    # finally xgboost
    risky_param_dicts = [
        {
            "max_depth": 5,  # > 3.5
            "n_estimators": 10,  #  and 3.5 < n_estimators <= 12.5
            "min_child_weight": 1,  #   and model.min_child_weight <= 1.5
        },
        {
            "max_depth": 5,  # > 3.5
            "n_estimators": 25,  # > 12.5
            "min_child_weight": 1,  # <= 3
        },
        {
            "max_depth": 5,  # > 3.5
            "n_estimators": 100,  # > 62.5
            "min_child_weight": 5,  #    and 3 < model.min_child_weight <= 6
        },
    ]
    for idx, paramdict in enumerate(risky_param_dicts):
        model = XGBClassifier(**paramdict)
        errstr = f" unnecessary risk with rule {idx}" f"params are {model.get_params()}"
        assert sa.get_unnecessary_risk(model) == 1, errstr
    model = XGBClassifier(min_child_weight=10)
    assert (
        sa.get_unnecessary_risk(model) == 0
    ), f"should be non-disclosive with {model.get_params}"


def test_non_trees():
    """Test  behaviour if model type not tree-based."""
    param_dict = {"probability": True}
    target = get_target("svc", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    # remove model
    target.model = None
    with pytest.raises(NotImplementedError):
        myattack2 = sa.StructuralAttack()
        myattack2.attack(target)


def test_dt():
    """Test for decision tree classifier."""

    # 'non' disclosive'
    param_dict = {"max_depth": 1, "min_samples_leaf": 150}
    target = get_target("dt", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    assert myattack.DoF_risk == 0, "should be no DoF risk with decision stump"
    assert (
        myattack.k_anonymity_risk == 0
    ), "should be no k-anonymity risk with min_samples_leaf 150"
    assert (
        myattack.class_disclosure_risk == 0
    ), "no class disclosure risk for stump with min samples leaf 150"
    assert myattack.unnecessary_risk == 0, "not unnecessary risk if max_depth < 3.5"

    # highly disclosive
    param_dict2 = {"max_depth": None, "min_samples_leaf": 1, "min_samples_split": 2}
    target = get_target("dt", **param_dict2)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    assert myattack.DoF_risk == 0, "should be no DoF risk with decision stump"
    assert (
        myattack.k_anonymity_risk == 1
    ), "should be  k-anonymity risk with unlimited depth and min_samples_leaf 5"
    assert (
        myattack.class_disclosure_risk == 1
    ), "should be class disclosure risk with unlimited depth and min_samples_leaf 5"
    assert (
        myattack.unnecessary_risk == 1
    ), " unnecessary risk with unlimited depth and min_samples_leaf 5"


def test_adaboost():
    """Test for adaboost classifier."""

    # 'non' disclosive'
    # - base estimator =None => DecisionTreeClassifier with max_depth 1
    # also set THRESHOLD to 4
    param_dict = {"n_estimators": 2, "base_estimator": None}
    target = get_target("adaboost", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.THRESHOLD = 2
    myattack.attack(target)
    assert myattack.DoF_risk == 0, "should be no DoF risk with just 2 decision stumps"
    assert (
        myattack.k_anonymity_risk == 0
    ), "should be no k-anonymity risk with only 2 stumps"
    assert myattack.class_disclosure_risk == 0, "no class disclosure risk for 2 stumps"
    assert myattack.unnecessary_risk == 0, " unnecessary risk not defined for adaboost"

    # highly disclosive
    kwargs = {"max_depth": None, "min_samples_leaf": 2}
    param_dict2 = {
        "base_estimator": DecisionTreeClassifier(**kwargs),
        "n_estimators": 1000,
    }
    target = get_target("adaboost", **param_dict2)
    myattack2 = sa.StructuralAttack()
    myattack2.attack(target)
    assert myattack2.DoF_risk == 1, "should be  DoF risk with adaboost of deep trees"
    assert (
        myattack2.k_anonymity_risk == 1
    ), "should be  k-anonymity risk with adaboost unlimited depth and min_samples_leaf 2"
    assert (
        myattack2.class_disclosure_risk == 1
    ), "should be class disclosure risk with adaboost unlimited depth and min_samples_leaf 2"
    assert myattack2.unnecessary_risk == 0, " unnecessary risk not define for adaboost"


def test_rf():
    """Test for random forest classifier."""

    # 'non' disclosive'
    param_dict = {"max_depth": 1, "min_samples_leaf": 150, "n_estimators": 10}
    target = get_target("rf", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    assert (
        myattack.DoF_risk == 0
    ), "should be no DoF risk with small forest of decision stumps"
    assert (
        myattack.k_anonymity_risk == 0
    ), "should be no k-anonymity risk with min_samples_leaf 150"
    assert (
        myattack.class_disclosure_risk == 0
    ), "no class disclosure risk for stumps with min samples leaf 150"
    assert myattack.unnecessary_risk == 0, "not unnecessary risk if max_depth < 3.5"

    # highly disclosive
    param_dict2 = {
        "max_depth": None,
        "min_samples_leaf": 2,
        "min_samples_split": 2,
        "n_estimators": 1000,
    }
    target = get_target("rf", **param_dict2)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    assert myattack.DoF_risk == 1, "should be  DoF risk with forest of deep trees"
    assert (
        myattack.k_anonymity_risk == 1
    ), "should be  k-anonymity risk with unlimited depth and min_samples_leaf 5"
    assert (
        myattack.class_disclosure_risk == 1
    ), "should be class disclsoure risk with unlimited depth and min_samples_leaf 5"
    assert (
        myattack.unnecessary_risk == 1
    ), " unnecessary risk with unlimited depth and min_samples_leaf 5"


def test_xgb():
    """Test for xgboost."""
    # non-disclosive
    param_dict = {"max_depth": 1, "min_child_weight": 50, "n_estimators": 5}
    target = get_target("xgb", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    assert (
        myattack.DoF_risk == 0
    ), "should be no DoF risk with small xgb of decision stumps"
    assert (
        myattack.k_anonymity_risk == 0
    ), "should be no k-anonymity risk with min_samples_leaf 150"
    assert (
        myattack.class_disclosure_risk == 0
    ), "no class disclosure risk for stumps with min child weight 50"
    assert myattack.unnecessary_risk == 0, "not unnecessary risk if max_depth < 3.5"

    # highly disclosive
    param_dict2 = {"max_depth": 50, "n_estimators": 100, "min_child_weight": 1}
    target2 = get_target("xgb", **param_dict2)
    myattack2 = sa.StructuralAttack()
    myattack2.attack(target2)
    assert myattack2.DoF_risk == 1, "should be  DoF risk with xgb of deep trees"
    assert (
        myattack2.k_anonymity_risk == 1
    ), "should be  k-anonymity risk with depth 50 and min_child_weight 1"
    assert (
        myattack2.class_disclosure_risk == 1
    ), "should be class disclosure risk with xgb lots of deep trees"
    assert myattack2.unnecessary_risk == 1, " unnecessary risk with these xgb params"


def test_reporting():
    """Test reporting functionality."""
    param_dict = {"max_depth": 1, "min_samples_leaf": 150}
    target = get_target("dt", **param_dict)
    myattack = sa.StructuralAttack()
    myattack.attack(target)
    myattack.make_report()


def test_main_example():
    """Test command line example."""
    param_dict = {"max_depth": 1, "min_samples_leaf": 150}
    target = get_target("dt", **param_dict)
    target_path = "dt.sav"
    target.save(target_path)
    testargs = [
        "prog",
        "run-attack",
        "--target-path",
        target_path,
        "--output-dir",
        "test_output_sa",
        "--report-name",
        "commandline_structural_report",
    ]
    with patch.object(sys, "argv", testargs):
        sa.main()
    config = {
        "output_dir": "test_output_structural2",
        "report_name": "structural_test",
    }
    with open("config_structural_test.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(config))
    testargs = [
        "prog",
        "run-attack-from-configfile",
        "--attack-config-json-file-name",
        "config_structural_test.json",
        "--attack-target-folder-path",
        "dt.sav",
    ]
    with patch.object(sys, "argv", testargs):
        sa.main()

    clean_up("dt.sav")
    clean_up("test_output_sa")
    clean_up("config_structural_test.json")



"""
Test to run multiple attacks (MIA and AIA) using a single configuration file
having different configuration settings (i.e. attack type or configuration parameters).

Running
-------

Invoke this code from the root AI-SDC folder.
However to run this test file, it will be required to install pytest package
using 'pip install pytest' and then run following
python -m pytest .\tests\test_multiple_attacks.py
"""

from __future__ import annotations

import json
import os
import shutil
import sys

# ignore unused imports because it depends on whether data file is present
from sklearn.datasets import fetch_openml  # pylint:disable=unused-import
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import (  # pylint:disable=unused-import
    LabelEncoder,
    OneHotEncoder,
)

from aisdc.attacks.multiple_attacks import (  # pylint: disable = import-error
    ConfigFile,
    MultipleAttacks,
)
from tests.test_attacks_via_safemodel import get_target

# pylint: disable = duplicate-code


def cleanup_file(name: str):
    """Removes unwanted files or directory."""
    if os.path.exists(name):
        if os.path.isfile(name):
            os.remove(name)
        elif os.path.isdir(name):
            shutil.rmtree(name)


def common_setup():
    """Basic commands to get ready to test some code."""
    model = RandomForestClassifier(bootstrap=False)
    target = get_target(model)
    model.fit(target.x_train, target.y_train)
    attack_obj = MultipleAttacks(
        config_filename="test_single_config.json",
    )
    return target, attack_obj


def create_single_config_file():
    """Creates single configuration file using multiple attack configuration."""
    # instantiating a configfile object to add configurations
    configfile_obj = ConfigFile(
        filename="test_single_config.json",
    )

    # Example 1: Adding three different worst case configuration dictionaries to the JSON file
    config = {
        "n_reps": 10,
        "n_dummy_reps": 1,
        "p_thresh": 0.05,
        "test_prop": 0.5,
        "train_beta": 5,
        "test_beta": 2,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
    }
    configfile_obj.add_config(config, "worst_case")

    config = {
        "n_reps": 20,
        "n_dummy_reps": 1,
        "p_thresh": 0.05,
        "test_prop": 0.5,
        "train_beta": 5,
        "test_beta": 2,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
    }
    configfile_obj.add_config(config, "worst_case")

    config = {
        "n_reps": 10,
        "n_dummy_reps": 1,
        "p_thresh": 0.05,
        "test_prop": 0.5,
        "train_beta": 5,
        "test_beta": 2,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
        "training_preds_filename": "train_preds.csv",
        "test_preds_filename": "test_preds.csv",
        "attack_metric_success_name": "P_HIGHER_AUC",
        "attack_metric_success_thresh": 0.05,
        "attack_metric_success_comp_type": "lte",
        "attack_metric_success_count_thresh": 2,
        "attack_fail_fast": True,
    }
    configfile_obj.add_config(config, "worst_case")

    # Adding two different lira attack configuration dictionaries to the JSON file
    config = {
        "n_shadow_models": 100,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
        "training_data_filename": "train_data.csv",
        "test_data_filename": "test_data.csv",
        "training_preds_filename": "train_preds.csv",
        "test_preds_filename": "test_preds.csv",
        "target_model": ["sklearn.ensemble", "RandomForestClassifier"],
        "target_model_hyp": {"min_samples_split": 2, "min_samples_leaf": 1},
    }
    configfile_obj.add_config(config, "lira")

    config = {
        "n_shadow_models": 150,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
        "shadow_models_fail_fast": True,
        "n_shadow_rows_confidences_min": 10,
        "training_data_filename": "train_data.csv",
        "test_data_filename": "test_data.csv",
        "training_preds_filename": "train_preds.csv",
        "test_preds_filename": "test_preds.csv",
        "target_model": ["sklearn.ensemble", "RandomForestClassifier"],
        "target_model_hyp": {"min_samples_split": 2, "min_samples_leaf": 1},
    }
    configfile_obj.add_config(config, "lira")
    # adding explicitly wrong attack name to cover codecov test
    configfile_obj.add_config(config, "lirrra")

    # Example 3: Adding a lira JSON configuration file to a configuration file
    # having multiple attack configurations
    config = {
        "n_shadow_models": 120,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
        "shadow_models_fail_fast": True,
        "n_shadow_rows_confidences_min": 10,
        "training_data_filename": "train_data.csv",
        "test_data_filename": "test_data.csv",
        "training_preds_filename": "train_preds.csv",
        "test_preds_filename": "test_preds.csv",
        "target_model": ["sklearn.ensemble", "RandomForestClassifier"],
        "target_model_hyp": {"min_samples_split": 2, "min_samples_leaf": 1},
    }
    with open("test_lira_config.json", "w", encoding="utf-8") as f:
        f.write(json.dumps(config))
    configfile_obj.add_config("test_lira_config.json", "lira")

    # Example 4: Adding a attribute configuration dictionary
    # from an existing configuration file to the JSON configuration file
    config = {
        "n_cpu": 2,
        "output_dir": "outputs_multiple_attacks",
        "report_name": "report_multiple_attacks",
    }
    configfile_obj.add_config(config, "attribute")
    os.remove("test_lira_config.json")
    return configfile_obj


def test_configfile_number():
    """Tests number of attack configurations in a configuration file."""
    configfile_obj = create_single_config_file()
    configfile_data = configfile_obj.read_config_file()
    assert len(configfile_data) == 8
    os.remove("test_single_config.json")


def test_multiple_attacks_programmatic():
    """Tests programmatically running attacks using a single configuration configuration file."""
    target, attack_obj = common_setup()
    _ = create_single_config_file()
    attack_obj.attack(target)
    print(attack_obj)
    os.remove("test_single_config.json")


def test_multiple_attacks_cmd():
    """Tests running multiple attacks (MIA and AIA) on the nursery data
    with an added continuous feature.
    """
    target, _ = common_setup()
    target.save(path=os.path.join("tests", "test_multiple_target"))
    _ = create_single_config_file()

    multiple_target = os.path.join("tests", "test_multiple_target")
    os.system(
        f"{sys.executable} -m aisdc.attacks.multiple_attacks run-attack-from-configfile "
        "--attack-config-json-file-name test_single_config.json "
        f"--attack-target-folder-path {multiple_target} "
    )


def test_cleanup():
    """Tidies up any files created."""
    files_made = (
        "test_single_config.json",
        "outputs_multiple_attacks",
        os.path.join("tests", "test_multiple_target"),
    )
    for fname in files_made:
        cleanup_file(fname)
